# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""This module provides a command-line interface for interacting with the Tesseract runtime."""

import contextlib
import io
import os
import sys
from collections.abc import Callable, Generator
from pathlib import Path
from textwrap import dedent
from typing import Any, Optional

import click
import typer
from pydantic import ValidationError

from tesseract_core.runtime.config import get_config, update_config
from tesseract_core.runtime.core import (
    check_tesseract_api,
    create_endpoints,
    get_tesseract_api,
)
from tesseract_core.runtime.file_interactions import (
    SUPPORTED_FORMATS,
    guess_format_from_path,
    load_bytes,
    output_to_bytes,
    read_from_path,
    write_to_path,
)
from tesseract_core.runtime.finite_differences import (
    check_gradients as check_gradients_,
)
from tesseract_core.runtime.serve import create_rest_api
from tesseract_core.runtime.serve import serve as serve_


class SpellcheckedTyperGroup(typer.core.TyperGroup):
    """A Typer group that suggests similar commands if a command is not found."""

    def get_command(self, ctx: click.Context, invoked_command: str) -> Any:
        """Get a command from the Typer group, suggesting similar commands if the command is not found."""
        import difflib

        possible_commands = self.list_commands(ctx)
        if invoked_command not in possible_commands:
            close_match = difflib.get_close_matches(
                invoked_command, possible_commands, n=1, cutoff=0.6
            )
            if close_match:
                raise click.UsageError(
                    f"No such command '{invoked_command}'. Did you mean '{close_match[0]}'?",
                    ctx,
                )
        return super().get_command(ctx, invoked_command)


app = typer.Typer(name="tesseract-runtime", cls=SpellcheckedTyperGroup)


def _prettify_docstring(docstr: str) -> str:
    """Enforce consistent indentation level of docstrings."""
    # First line is not indented, the rest is -> leads to formatting issues
    docstring_lines = docstr.split("\n")
    dedented_lines = dedent("\n".join(docstring_lines[1:]))
    return "\n".join([docstring_lines[0].lstrip(), dedented_lines])


def _parse_arg_callback(
    ctx: Any, param: Any, value: Any
) -> tuple[dict[str, Any], Optional[Path]]:
    """Click callback to parse Tesseract input arguments provided in the CLI.

    Returns a tuple of the parsed value and the base directory of the input path, if any.
    """
    base_dir = None

    if not isinstance(value, str):
        # Passthrough, probably a default value
        return value, base_dir

    if value.startswith("@"):
        base_dir = Path(value[1:]).parent
        value_format = guess_format_from_path(value[1:])
        try:
            value_bytes = read_from_path(value[1:])
        except Exception as e:
            raise click.BadParameter(f"Could not read data from path {value}.") from e
    else:
        # Data given directly via the CLI is always in JSON format
        value_format = "json"
        value_bytes = value.encode()

    try:
        decoded_value = load_bytes(value_bytes, value_format)
    except Exception as e:
        raise click.BadParameter(
            f"Could not decode data using format {value_format}"
        ) from e

    return decoded_value, base_dir


@app.callback(
    no_args_is_help=False,
    context_settings={"help_option_names": ["-h", "--help"]},
)
def main_callback() -> None:
    """Invoke the Tesseract runtime.

    The Tesseract runtime can be configured via environment variables; for example,
    ``export TESSERACT_RUNTIME_PORT=8080`` sets the port to use for ``tesseract serve`` to 8080.
    """
    pass


tesseract_runtime = typer.main.get_command(app)


def _schema_to_docstring(schema: Any, current_indent: int = 0) -> str:
    """Convert a Pydantic schema to a human-readable docstring."""
    docstring = []
    indent = " " * current_indent

    if not hasattr(schema, "model_fields"):
        return ""

    is_root = schema.model_fields.keys() == {"root"}

    for field, field_info in schema.model_fields.items():
        if is_root:
            docline = f"{indent}{field_info.description}"
        else:
            docline = f"{indent}{field}: {field_info.description}"

        if (
            field_info.default is not None
            and str(field_info.default) != "PydanticUndefined"
        ):
            docline = f"{docline} [default: {field_info.default}]"

        docstring.append(docline)

        if hasattr(field_info.annotation, "model_fields"):
            docstring.append(
                _schema_to_docstring(field_info.annotation, current_indent + 4)
            )

    return "\n".join(docstring)


@tesseract_runtime.command()
def check() -> None:
    """Check whether the Tesseract API is valid."""
    api_module = get_tesseract_api()
    check_tesseract_api(api_module)
    typer.echo("✅ Tesseract API check successful ✅")


@tesseract_runtime.command()
@click.argument(
    "payload",
    type=click.STRING,
    required=True,
    metavar="JSON_PAYLOAD",
    callback=_parse_arg_callback,
)
@click.option(
    "--endpoints",
    type=click.STRING,
    required=False,
    multiple=True,
    help="Endpoints to check gradients for (default: check all).",
)
@click.option(
    "--input-paths",
    type=click.STRING,
    required=False,
    multiple=True,
    help="Paths to differentiable inputs to check gradients for (default: check all).",
)
@click.option(
    "--output-paths",
    type=click.STRING,
    required=False,
    multiple=True,
    help="Paths to differentiable outputs to check gradients for (default: check all).",
)
@click.option(
    "--eps",
    type=click.FLOAT,
    required=False,
    help="Step size for finite differences.",
    default=1e-4,
    show_default=True,
)
@click.option(
    "--rtol",
    type=click.FLOAT,
    required=False,
    help="Relative tolerance when comparing finite differences to gradients.",
    default=0.1,
    show_default=True,
)
@click.option(
    "--max-evals",
    type=click.INT,
    required=False,
    help="Maximum number of evaluations per input.",
    default=1000,
    show_default=True,
)
@click.option(
    "--max-failures",
    type=click.INT,
    required=False,
    help="Maximum number of failures to report per endpoint.",
    default=10,
    show_default=True,
)
@click.option(
    "--seed",
    type=click.INT,
    required=False,
    help="Seed for random number generator. If not set, a random seed is used.",
    default=None,
)
@click.option(
    "--show-progress",
    is_flag=True,
    default=True,
    help="Show progress bar.",
)
def check_gradients(
    payload: tuple[dict[str, Any], Optional[Path]],
    input_paths: list[str],
    output_paths: list[str],
    endpoints: list[str],
    eps: float,
    rtol: float,
    max_evals: int,
    max_failures: int,
    seed: Optional[int],
    show_progress: bool,
) -> None:
    """Check gradients of endpoints against a finite difference approximation.

    This is an automated way to check the correctness of the gradients of the different AD endpoints
    (jacobian, jacobian_vector_product, vector_jacobian_product) of a ``tesseract_api.py`` module.
    It will sample random indices and compare the gradients computed by the AD endpoints with the
    finite difference approximation.

    Warning:
        Finite differences are not exact and the comparison is done with a tolerance. This means
        that the check may fail even if the gradients are correct, and vice versa.

    Finite difference approximations are sensitive to numerical precision. When finite differences
    are reported incorrectly as 0.0, it is likely that the chosen `eps` is too small, especially for
    inputs that do not use float64 precision.
    """
    api_module = get_tesseract_api()
    inputs, base_dir = payload

    result_iter = check_gradients_(
        api_module,
        inputs,
        base_dir=base_dir,
        input_paths=input_paths,
        output_paths=output_paths,
        endpoints=endpoints,
        max_evals=max_evals,
        eps=eps,
        rtol=rtol,
        seed=seed,
        show_progress=show_progress,
    )

    failed = False
    for endpoint, failures, num_evals in result_iter:
        if not failures:
            typer.echo(
                f"✅ Gradient check for {endpoint} passed ✅ ({len(failures)} failures / {num_evals} checks)"
            )
        else:
            failed = True
            typer.echo()
            typer.echo(
                f"⚠️ Gradient check for {endpoint} failed ⚠️ ({len(failures)} failures / {num_evals} checks)"
            )
            printed_failures = min(len(failures), max_failures)
            typer.echo(f"First {printed_failures} failures:")
            for failure in failures[:printed_failures]:
                typer.echo(
                    f"  Input path: '{failure.in_path}', Output path: '{failure.out_path}', Index: {failure.idx}"
                )
                if failure.exception:
                    typer.echo(f"  Encountered exception: {failure.exception}")
                else:
                    typer.echo(f"  {endpoint} value: {failure.grad_val}")
                    typer.echo(f"  Finite difference value: {failure.ref_val}")
                typer.echo()

    if failed:
        typer.echo("❌ Some gradient checks failed ❌")
        sys.exit(1)


@tesseract_runtime.command()
@click.option("-p", "--port", default=8000, help="Port number")
@click.option("-h", "--host", default="127.0.0.1", help="Host IP address")
@click.option("-w", "--num-workers", default=1, help="Number of worker processes")
def serve(host: str, port: int, num_workers: int) -> None:
    """Start running this Tesseract's web server."""
    serve_(host=host, port=port, num_workers=num_workers)


def _create_user_defined_cli_command(
    user_function: Callable, out_stream: Optional[io.IOBase]
) -> click.Command:
    """Creates a click command which sends requests to Tesseract endpoints.

    We need to do this dynamically, as we want to generate docs and usage
    from the Tesseract api's signature and docstrings.

    Args:
        user_function: The user-defined function to create a CLI command for.
        out_stream: The default output stream to write to. If None, defaults to
            sys.stdout at the time of invocation.
    """
    InputSchema = user_function.__annotations__.get("payload", None)
    OutputSchema = user_function.__annotations__.get("return", None)

    options = []

    if InputSchema is not None:
        options.append(
            click.Argument(
                ["payload"],
                type=click.STRING,
                required=True,
                metavar="JSON_PAYLOAD",
                callback=_parse_arg_callback,
            )
        )

    options.extend(
        [
            click.Option(
                ["-i", "--input-path"],
                type=click.STRING,
                help=(
                    "Input path from which to read files. "
                    "To be used in conjunction with InputFileReferences in the Tesseract InputSchema."
                ),
                default=None,
            ),
            click.Option(
                ["-o", "--output-path"],
                type=click.STRING,
                help=(
                    "Output path to write the result to, such as local directory or S3 URI "
                    "(may be anything supported by fsspec). [default: write to stdout]"
                ),
                default=None,
            ),
            click.Option(
                ["-f", "--output-format"],
                type=click.Choice(SUPPORTED_FORMATS),
                help="Output format to write results in.",
                default="json",
            ),
        ]
    )

    def _callback_wrapper(
        input_path: Optional[str],
        output_path: Optional[str],
        output_format: SUPPORTED_FORMATS,
        **optional_args: Any,
    ):
        if input_path:
            update_config(input_path=input_path)

        if output_format == "json+binref" and output_path is None:
            raise ValueError("--output-path must be specified for json+binref format")

        out_stream_ = out_stream or sys.stdout

        user_function_args = {}

        if InputSchema is not None:
            payload, base_dir = optional_args["payload"]
            try:
                user_function_args["payload"] = InputSchema.model_validate(
                    payload, context={"base_dir": base_dir}
                )
            except ValidationError as e:
                raise click.BadParameter(
                    e,
                    param=InputSchema,
                    param_hint=InputSchema.__name__,
                ) from e

        if output_path:
            update_config(output_path=output_path)
            Path(output_path).mkdir(parents=True, exist_ok=True)

        result = user_function(**user_function_args)
        result = output_to_bytes(result, output_format, output_path)

        if output_path is None:
            # write raw bytes to out_stream.buffer to support binary data (which may e.g. be piped)
            out_stream_.buffer.write(result)
            out_stream_.flush()
        else:
            format = output_format.split("+", maxsplit=1)[0]
            write_to_path(result, f"{output_path}/results.{format}")

    function_name = user_function.__name__.replace("_", "-")

    # Assemble docstring
    function_doc = [_prettify_docstring(user_function.__doc__)]

    if InputSchema is not None and hasattr(InputSchema, "model_fields"):
        function_doc.append(
            "\nFirst argument is the payload, which should be a JSON object with the following structure."
        )

        # \b\n disables click's automatic formatting
        function_doc.append("\n\n\b\nInput schema:")
        function_doc.append(_schema_to_docstring(InputSchema, current_indent=4))

    if OutputSchema is not None and hasattr(OutputSchema, "model_fields"):
        function_doc.append("\n\n\b\nReturns:")
        function_doc.append(_schema_to_docstring(OutputSchema, current_indent=4))

    command = click.Command(
        function_name,
        short_help=f"Call the Tesseract function {function_name}.",
        help="\n".join(function_doc),
        callback=_callback_wrapper,
        params=options,
    )

    return command


def _add_user_commands_to_cli(
    group: click.Group, out_stream: Optional[io.IOBase]
) -> click.Group:
    tesseract_package = get_tesseract_api()
    endpoints = create_endpoints(tesseract_package)

    openapi_schema_ = create_rest_api(tesseract_package).openapi()

    def openapi_schema() -> dict:
        """Get the openapi.json schema."""
        return openapi_schema_

    endpoints.append(openapi_schema)

    for func in endpoints:
        group.add_command(_create_user_defined_cli_command(func, out_stream))

    return group


@contextlib.contextmanager
def stdout_to_stderr() -> Generator:
    """Redirect stdout to stderr at OS level."""
    orig_stdout = os.dup(sys.stdout.fileno())
    sys.stdout.flush()
    os.dup2(sys.stderr.fileno(), sys.stdout.fileno())
    try:
        yield os.fdopen(orig_stdout, "w", closefd=False)
    finally:
        sys.stdout.flush()
        os.dup2(orig_stdout, sys.stdout.fileno())


def main() -> None:
    """Entrypoint for the command line interface."""
    # Redirect stdout to stderr to avoid mixing any output with the JSON response.
    with stdout_to_stderr() as orig_stdout:
        # Fail as fast as possible if the Tesseract API path is not set
        api_path = get_config().api_path
        if not api_path.is_file():
            print(
                f"Tesseract API file '{api_path}' does not exist. "
                "Please ensure it is a valid file, or set the TESSERACT_API_PATH "
                "environment variable to the path of your Tesseract API file.\n"
                "\n"
                "Example:\n"
                "    $ export TESSERACT_API_PATH=/path/to/your/tesseract_api.py\n"
                "\n"
                "Aborted.",
                file=sys.stderr,
            )
            sys.exit(1)

        cli = _add_user_commands_to_cli(tesseract_runtime, out_stream=orig_stdout)
        cli(auto_envvar_prefix="TESSERACT_RUNTIME")


if __name__ == "__main__":
    main()



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

# Tesseract API module for lorenz
# Generated by tesseract 0.8.5 on 2025-04-30T09:44:04.234338

from typing import Any

import equinox as eqx
import jax
import jax.numpy as jnp
from pydantic import BaseModel, Field

from tesseract_core.runtime import Array, Differentiable, Float32
from tesseract_core.runtime.tree_transforms import filter_func, flatten_with_paths


class InputSchema(BaseModel):
    """Input schema for forecasting of Lorenz 96 system."""

    state: Differentiable[Array[(None,), Float32]] = Field(
        description="A state vector for the Lorenz 96 system"
    )
    F: float = Field(description="Forcing parameter for Lorenz 96", default=8.0)
    dt: float = Field(description="Time step for integration", default=0.05)
    n_steps: int = Field(description="Number of integration steps", default=1)


class OutputSchema(BaseModel):
    """Output schema for forecasting of Lorenz 96 system."""

    result: Differentiable[Array[(None, None), Float32]] = Field(
        description="A trajectorie of predictions after integration"
    )


def lorenz96_step(state: jnp.ndarray, F: float, dt: float) -> jnp.ndarray:
    """Perform one step of RK4 integration for the Lorenz 96 system."""

    def lorenz96_derivatives(x: jnp.ndarray) -> jnp.ndarray:
        """Compute the derivatives for Lorenz 96 system."""
        N = x.shape[0]
        # Create arrays for indices with wraparound
        ip1 = (jnp.arange(N) + 1) % N  # i+1 with wraparound
        im1 = (jnp.arange(N) - 1) % N  # i-1 with wraparound
        im2 = (jnp.arange(N) - 2) % N  # i-2 with wraparound
        # Compute derivatives: dx_i/dt = (x_{i+1} - x_{i-2}) * x_{i-1} - x_i + F
        d = (x[ip1] - x[im2]) * x[im1] - x + F
        return d

    # RK4 integration
    k1 = lorenz96_derivatives(state)
    k2 = lorenz96_derivatives(state + dt * k1 / 2)
    k3 = lorenz96_derivatives(state + dt * k2 / 2)
    k4 = lorenz96_derivatives(state + dt * k3)
    return state + dt * (k1 + 2 * k2 + 2 * k3 + k4) / 6


def lorenz96_multi_step(
    state: jnp.ndarray, F: float, dt: float, n_steps: int
) -> jnp.ndarray:
    """Perform multiple steps of Lorenz 96 integration using scan."""

    def step_fn(state: jnp.ndarray, _: Any) -> tuple:
        return lorenz96_step(state, F, dt), state

    _, trajectory = jax.lax.scan(step_fn, state, None, length=n_steps)
    return trajectory


@eqx.filter_jit
def apply_jit(inputs: dict) -> dict:
    """The apply_jit function for the Lorenz 96 tesseract."""
    trajectory = lorenz96_multi_step(**inputs)
    return dict(result=trajectory)


def apply(inputs: InputSchema) -> OutputSchema:
    """The apply function for the Lorenz 96 tesseract."""
    return apply_jit(inputs.model_dump())


def abstract_eval(abstract_inputs: Any) -> Any:
    """Calculate output shape of apply from the shape of its inputs."""
    is_shapedtype_dict = lambda x: type(x) is dict and (x.keys() == {"shape", "dtype"})
    is_shapedtype_struct = lambda x: isinstance(x, jax.ShapeDtypeStruct)

    jaxified_inputs = jax.tree.map(
        lambda x: jax.ShapeDtypeStruct(**x) if is_shapedtype_dict(x) else x,
        abstract_inputs.model_dump(),
        is_leaf=is_shapedtype_dict,
    )
    dynamic_inputs, static_inputs = eqx.partition(
        jaxified_inputs, filter_spec=is_shapedtype_struct
    )

    def wrapped_apply(dynamic_inputs: Any) -> Any:
        inputs = eqx.combine(static_inputs, dynamic_inputs)
        return apply_jit(inputs)

    jax_shapes = jax.eval_shape(wrapped_apply, dynamic_inputs)
    return jax.tree.map(
        lambda x: (
            {"shape": x.shape, "dtype": str(x.dtype)} if is_shapedtype_struct(x) else x
        ),
        jax_shapes,
        is_leaf=is_shapedtype_struct,
    )


@eqx.filter_jit
def jac_jit(
    inputs: dict,
    jac_inputs: tuple[str],
    jac_outputs: tuple[str],
) -> dict:
    """The jac_jit function for the Lorenz 96 tesseract."""
    filtered_apply = filter_func(apply_jit, inputs, jac_outputs)
    return jax.jacrev(filtered_apply)(
        flatten_with_paths(inputs, include_paths=jac_inputs)
    )


def jacobian(
    inputs: InputSchema,
    jac_inputs: set[str],
    jac_outputs: set[str],
) -> Any:
    """The jacobian function for the Lorenz 96 tesseract."""
    return jac_jit(inputs.model_dump(), tuple(jac_inputs), tuple(jac_outputs))


@eqx.filter_jit
def jvp_jit(
    inputs: dict,
    jvp_inputs: tuple[str],
    jvp_outputs: tuple[str],
    tangent_vector: dict,
) -> Any:
    """The jvp_jit function for the Lorenz 96 tesseract."""
    filtered_apply = filter_func(apply_jit, inputs, jvp_outputs)
    return jax.jvp(
        filtered_apply,
        [flatten_with_paths(inputs, include_paths=jvp_inputs)],
        [tangent_vector],
    )[1]


def jacobian_vector_product(
    inputs: InputSchema,
    jvp_inputs: set[str],
    jvp_outputs: set[str],
    tangent_vector: dict[str, Any],
) -> Any:
    """The jacobian vector product function for the Lorenz 96 tesseract."""
    return jvp_jit(
        inputs.model_dump(),
        tuple(jvp_inputs),
        tuple(jvp_outputs),
        tangent_vector,
    )


@eqx.filter_jit
def vjp_jit(
    inputs: dict,
    vjp_inputs: tuple[str],
    vjp_outputs: tuple[str],
    cotangent_vector: dict,
) -> Any:
    """The vjp_jit function for the Lorenz 96 tesseract."""
    filtered_apply = filter_func(apply_jit, inputs, vjp_outputs)
    _, vjp_func = jax.vjp(
        filtered_apply, flatten_with_paths(inputs, include_paths=vjp_inputs)
    )
    return vjp_func(cotangent_vector)[0]


def vector_jacobian_product(
    inputs: InputSchema,
    vjp_inputs: set[str],
    vjp_outputs: set[str],
    cotangent_vector: dict[str, Any],
) -> Any:
    """The vector jacobian product function for the Lorenz 96 tesseract."""
    return vjp_jit(
        inputs.model_dump(),
        tuple(vjp_inputs),
        tuple(vjp_outputs),
        cotangent_vector,
    )



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""Metrics, Parameters, and Artifacts (MPA) library for Tesseract Core."""

import csv
import json
import os
import shutil
from abc import ABC, abstractmethod
from collections.abc import Generator
from contextlib import contextmanager
from contextvars import ContextVar
from datetime import datetime
from pathlib import Path
from typing import Any, Optional

import requests

from tesseract_core.runtime.config import get_config


class BaseBackend(ABC):
    """Base class for MPA backends."""

    @abstractmethod
    def log_parameter(self, key: str, value: Any) -> None:
        """Log a parameter."""
        pass

    @abstractmethod
    def log_metric(self, key: str, value: float, step: Optional[int] = None) -> None:
        """Log a metric."""
        pass

    @abstractmethod
    def log_artifact(self, local_path: str) -> None:
        """Log an artifact."""
        pass

    @abstractmethod
    def start_run(self) -> None:
        """Start a new run."""
        pass

    @abstractmethod
    def end_run(self) -> None:
        """End the current run."""
        pass


class FileBackend(BaseBackend):
    """MPA backend that writes to local files."""

    def __init__(self) -> None:
        self.log_dir = os.getenv("MPA_DIR")
        if not self.log_dir:
            self.log_dir = Path(get_config().output_path) / "mpa"
        else:
            self.log_dir = Path(self.log_dir)
        self.log_dir.mkdir(exist_ok=True)

        # Create a unique run directory
        timestamp = datetime.now().strftime("%Y%m%d_%H%M%S_%f")
        self.run_dir = self.log_dir / f"run_{timestamp}"
        self.run_dir.mkdir(exist_ok=True)

        # Initialize log files
        self.params_file = self.run_dir / "parameters.json"
        self.metrics_file = self.run_dir / "metrics.csv"
        self.artifacts_dir = self.run_dir / "artifacts"
        self.artifacts_dir.mkdir(exist_ok=True)

        # Initialize parameters dict and metrics list
        self.parameters = {}
        self.metrics = []

        # Initialize CSV file with headers
        with open(self.metrics_file, "w", newline="") as f:
            writer = csv.writer(f)
            writer.writerow(["timestamp", "key", "value", "step"])

    def log_parameter(self, key: str, value: Any) -> None:
        """Log a parameter to JSON file."""
        self.parameters[key] = value
        with open(self.params_file, "w") as f:
            json.dump(self.parameters, f, indent=2, default=str)

    def log_metric(self, key: str, value: float, step: Optional[int] = None) -> None:
        """Log a metric to CSV file."""
        timestamp = datetime.now().isoformat()
        step_value = (
            step
            if step is not None
            else len([m for m in self.metrics if m["key"] == key])
        )

        metric_entry = {
            "timestamp": timestamp,
            "key": key,
            "value": value,
            "step": step_value,
        }
        self.metrics.append(metric_entry)

        with open(self.metrics_file, "a", newline="") as f:
            writer = csv.writer(f)
            writer.writerow([timestamp, key, value, step_value])

    def log_artifact(self, local_path: str) -> None:
        """Copy artifact to the artifacts directory."""
        source_path = Path(local_path)
        if not source_path.exists():
            raise FileNotFoundError(f"Artifact file not found: {local_path}")

        dest_path = self.artifacts_dir / source_path.name
        shutil.copy2(source_path, dest_path)

    def start_run(self) -> None:
        """Start a new run. File backend doesn't need special start logic."""
        pass

    def end_run(self) -> None:
        """End the current run. File backend doesn't need special end logic."""
        pass


class MLflowBackend(BaseBackend):
    """MPA backend that writes to an MLflow tracking server."""

    def __init__(self) -> None:
        try:
            os.environ["GIT_PYTHON_REFRESH"] = (
                "quiet"  # Suppress potential MLflow git warnings
            )
            self._ensure_mlflow_reachable()

            import mlflow

            self.mlflow = mlflow
        except ImportError as exc:
            raise ImportError(
                "MLflow is required for MLflowBackend but is not installed"
            ) from exc

    def _ensure_mlflow_reachable(self) -> None:
        try:
            mlflow_tracking_uri = os.environ.get("MLFLOW_TRACKING_URI")
            if mlflow_tracking_uri and (
                mlflow_tracking_uri.startswith("http://")
                or mlflow_tracking_uri.startswith("https://")
            ):
                response = requests.get(mlflow_tracking_uri, timeout=5)
                response.raise_for_status()
        except requests.RequestException as e:
            raise RuntimeError(
                f"Failed to connect to MLflow tracking server at {mlflow_tracking_uri}. "
                "Please make sure an MLflow server is running and MLFLOW_TRACKING_URI is set correctly. "
                "Alternatively, switch to local file-based by setting MLFLOW_TRACKING_URI to an empty string."
            ) from e

    def log_parameter(self, key: str, value: Any) -> None:
        """Log a parameter to MLflow."""
        self.mlflow.log_param(key, value)

    def log_metric(self, key: str, value: float, step: Optional[int] = None) -> None:
        """Log a metric to MLflow."""
        self.mlflow.log_metric(key, value, step=step)

    def log_artifact(self, local_path: str) -> None:
        """Log an artifact to MLflow."""
        self.mlflow.log_artifact(local_path)

    def start_run(self) -> None:
        """Start a new MLflow run."""
        self.mlflow.start_run()

    def end_run(self) -> None:
        """End the current MLflow run."""
        self.mlflow.end_run()


def _create_backend() -> BaseBackend:
    """Create the appropriate backend based on environment."""
    if os.getenv("MLFLOW_TRACKING_URI"):
        return MLflowBackend()
    else:
        return FileBackend()


# Context variable for the current backend instance
_current_backend: ContextVar[BaseBackend] = ContextVar("current_backend")


def _get_current_backend() -> BaseBackend:
    """Get the current backend instance from context variable."""
    try:
        return _current_backend.get()
    except LookupError as exc:
        raise RuntimeError(
            "No active MPA run. Use 'with mpa.start_run():' to start a run."
        ) from exc


# Public API functions that work with the current context
def log_parameter(key: str, value: Any) -> None:
    """Log a parameter to the current run context."""
    _get_current_backend().log_parameter(key, value)


def log_metric(key: str, value: float, step: Optional[int] = None) -> None:
    """Log a metric to the current run context."""
    _get_current_backend().log_metric(key, value, step)


def log_artifact(local_path: str) -> None:
    """Log an artifact to the current run context."""
    _get_current_backend().log_artifact(local_path)


@contextmanager
def start_run() -> Generator[None, None, None]:
    """Context manager for starting and ending a run."""
    backend = _create_backend()
    token = _current_backend.set(backend)
    backend.start_run()
    try:
        yield
    finally:
        backend.end_run()
        _current_backend.reset(token)



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import copy
import json
import os
import subprocess
import sys
import traceback
from copy import deepcopy
from pathlib import Path

import boto3
import fsspec
import numpy as np
import pytest
from click.testing import CliRunner
from moto.server import ThreadedMotoServer

import tesseract_core.runtime
from tesseract_core.runtime.cli import _add_user_commands_to_cli
from tesseract_core.runtime.cli import tesseract_runtime as cli_cmd
from tesseract_core.runtime.file_interactions import load_bytes, output_to_bytes

test_input = {
    "a": [1.0, 2.0, 3.0],
    "b": [1, 1, 1],
    "s": 2.5,
}
test_input_binref = {
    "a": {
        "object_type": "array",
        "shape": [3],
        "dtype": "float64",
        "data": {
            "buffer": "data.bin",
            "encoding": "binref",
        },
    },
    "b": {
        "object_type": "array",
        "shape": [3],
        "dtype": "int64",
        "data": {
            "buffer": "data.bin:24",
            "encoding": "binref",
        },
    },
    "s": test_input["s"],
}


@pytest.fixture(autouse=True)
def use_dummy_tesseract(dummy_tesseract):
    yield


@pytest.fixture
def cli_runner():
    return CliRunner(mix_stderr=False)


@pytest.fixture(scope="module")
def test_s3_server():
    """Fixture to run a mocked AWS server for testing."""
    current_conf = copy.deepcopy(fsspec.config.conf)

    server = ThreadedMotoServer(port=0)
    server.start()

    os.environ.update(
        AWS_ACCESS_KEY_ID="dummy-access-key",
        AWS_SECRET_ACCESS_KEY="dummy-access-key-secret",
        AWS_DEFAULT_REGION="us-east-1",
    )

    try:
        host, port = server.get_host_and_port()

        moto_server = f"http://{host}:{port}"

        client = boto3.client("s3", endpoint_url=moto_server)
        client.create_bucket(Bucket="test-bucket")

        client.put_object(
            Bucket="test-bucket",
            Key="foo.json",
            Body=json.dumps({"inputs": test_input}),
        )

        for array in ("a", "b"):
            client.put_object(
                Bucket="test-bucket",
                Key=f"array_{array}.bin",
                Body=np.asarray(test_input[array]).tobytes(),
            )

        fsspec.config.conf["s3"] = dict(endpoint_url=moto_server)
        yield "s3://test-bucket"
    finally:
        server.stop()
        fsspec.config.conf = current_conf
        os.environ.pop("AWS_ACCESS_KEY_ID")
        os.environ.pop("AWS_SECRET_ACCESS_KEY")
        os.environ.pop("AWS_DEFAULT_REGION")


@pytest.fixture()
def test_http_server(free_port):
    """Start a local HTTP server for testing that always returns `test_input`."""
    import http.server
    import socketserver
    import threading

    import requests

    class Handler(http.server.SimpleHTTPRequestHandler):
        def do_GET(self):
            if "/foo.json" == self.path:
                bytes = json.dumps({"inputs": test_input}).encode("utf-8")
            elif "/array_a.bin" == self.path:
                bytes = np.asarray(test_input["a"]).tobytes()
            elif "/array_b.bin" == self.path:
                bytes = np.asarray(test_input["b"]).tobytes()
            else:
                raise ValueError(f"No such file {self.path}")
            self.send_response(200)
            self.end_headers()
            self.wfile.write(bytes)

    httpd = socketserver.TCPServer(("", free_port), Handler)

    def run_server():
        httpd.serve_forever()

    server_thread = threading.Thread(target=run_server, daemon=True)
    server_thread.start()

    try:
        # wait for server to start
        while True:
            res = requests.get(f"http://localhost:{free_port}/foo.json")
            if res.status_code == 200:
                break

        yield f"http://localhost:{free_port}"
    finally:
        httpd.shutdown()
        server_thread.join()


@pytest.fixture
def cli():
    new_cmd = copy.deepcopy(cli_cmd)
    return _add_user_commands_to_cli(new_cmd, out_stream=None)


def test_invocation_no_args_prints_usage(cli, cli_runner):
    result = cli_runner.invoke(cli, env={"NO_COLOR": "true"})
    assert result.exit_code == 2, result.stdout
    assert "Usage: tesseract-runtime" in result.stderr


def test_input_val_schema_command(cli, cli_runner):
    result = cli_runner.invoke(cli, ["input-schema"], catch_exceptions=False)
    assert result.exit_code == 0, result.stderr
    assert "properties" in result.stdout


def test_output_schema_command(cli, cli_runner):
    result = cli_runner.invoke(cli, ["output-schema"], catch_exceptions=False)
    assert result.exit_code == 0, result.stderr
    assert "properties" in result.stdout


def test_openapi_schema_command(cli, cli_runner):
    result = cli_runner.invoke(cli, ["openapi-schema"])
    assert result.exit_code == 0, result.stderr
    assert '"openapi":' in result.stdout


def test_abstract_eval_command(cli, cli_runner):
    input_shape = {"dtype": "float32", "shape": [4]}
    result = cli_runner.invoke(
        cli,
        [
            "abstract-eval",
            json.dumps({"inputs": {"a": input_shape, "b": input_shape}}),
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 0, result.stderr
    json_out = json.loads(result.stdout)
    assert {"result": {"shape": [4], "dtype": "float32"}} == json_out


def test_apply_command(cli, cli_runner, dummy_tesseract_module):
    inputs = dummy_tesseract_module.InputSchema(**test_input).model_dump_json(
        context={"array_encoding": "base64"}
    )
    inputs = json.loads(inputs)
    inputs = {"inputs": inputs}
    result = cli_runner.invoke(
        cli,
        ["apply", json.dumps(inputs)],
        catch_exceptions=False,
    )
    assert result.exit_code == 0, result.stderr
    test_input_val = dummy_tesseract_module.InputSchema.model_validate(test_input)
    expected = dummy_tesseract_module.apply(test_input_val).model_dump_json()
    assert json.loads(result.stdout) == json.loads(expected)


def test_apply_command_binref(cli, cli_runner, dummy_tesseract_module, tmpdir):
    # construct payload with absolute paths
    test_input_absolute = deepcopy(test_input_binref)
    test_input_absolute["a"]["data"]["buffer"] = str(tmpdir / "data.bin")
    test_input_absolute["b"]["data"]["buffer"] = str(tmpdir / "data.bin:24")

    # write array to current tempdir so that we can read from here
    with open(tmpdir / "data.bin", "wb") as fi:
        fi.write(np.asarray(test_input["a"]).tobytes())
        fi.write(np.asarray(test_input["b"]).tobytes())

    result = cli_runner.invoke(
        cli,
        ["apply", json.dumps({"inputs": test_input_absolute})],
        catch_exceptions=False,
    )
    assert result.exit_code == 0, result.stderr
    test_input_val = dummy_tesseract_module.InputSchema.model_validate(test_input)
    expected = dummy_tesseract_module.apply(test_input_val).model_dump_json()
    assert json.loads(result.stdout) == json.loads(expected)

    result = cli_runner.invoke(
        cli,
        ["apply", json.dumps({"inputs": test_input_binref})],
        catch_exceptions=False,
        env={"TERM": "true", "COLUMNS": "9999"},
    )
    assert result.exit_code == 2
    assert "Value error" in result.stderr
    assert "binref encoded with a relative path" in result.stderr


def test_apply_command_noenv(
    cli, cli_runner, dummy_tesseract_module, dummy_tesseract_noenv
):
    # ensure dummy_tesseract_noenv properly unset TESSERACT_API_PATH
    assert "TESSERACT_API_PATH" not in os.environ
    inputs = dummy_tesseract_module.InputSchema(**test_input).model_dump_json(
        context={"array_encoding": "base64"}
    )
    inputs = json.loads(inputs)
    inputs = {"inputs": inputs}
    result = cli_runner.invoke(
        cli,
        ["apply", json.dumps(inputs)],
        catch_exceptions=False,
    )
    assert result.exit_code == 0, result.stderr
    test_input_val = dummy_tesseract_module.InputSchema.model_validate(test_input)
    expected = dummy_tesseract_module.apply(test_input_val).model_dump_json()
    assert json.loads(result.stdout) == json.loads(expected)


@pytest.mark.parametrize(
    "input_format", ["json", "msgpack", "json+base64", "json+binref"]
)
def test_input_vals_from_local_file(
    cli, cli_runner, tmpdir, dummy_tesseract_module, input_format
):
    """Test the apply command with input arguments from a local file."""
    if "+" in input_format:
        container = input_format.split("+")[0]
    else:
        container = input_format

    a_file = tmpdir / f"a.{container}"
    inputs = {"inputs": dummy_tesseract_module.InputSchema(**test_input)}
    input_bytes = output_to_bytes(inputs, input_format, base_dir=Path(tmpdir))

    if input_format == "json+binref":
        # Make sure a binref file is created
        assert list(Path(tmpdir).glob("*.bin")) != []

    with open(a_file, "wb") as f:
        f.write(input_bytes)

    result = cli_runner.invoke(
        cli,
        ["apply", f"@{a_file}"],
        catch_exceptions=False,
    )
    assert result.exit_code == 0, result.stderr

    result = result.stdout

    test_input_val = dummy_tesseract_module.InputSchema.model_validate(test_input)
    expected = dummy_tesseract_module.apply(test_input_val).model_dump_json()
    assert json.loads(result) == json.loads(expected)


@pytest.mark.parametrize(
    "output_format", ["json", "msgpack", "json+base64", "json+binref"]
)
def test_outputs_to_local_file(
    cli, cli_runner, tmpdir, dummy_tesseract_module, output_format
):
    """Test the apply command writing outputs to a local file."""
    tmpdir = Path(tmpdir)
    result = cli_runner.invoke(
        cli,
        [
            "apply",
            json.dumps({"inputs": test_input}),
            "--output-path",
            tmpdir,
            "--output-format",
            output_format,
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 0, result.stderr

    test_input_val = dummy_tesseract_module.InputSchema.model_validate(test_input)

    container = output_format.split("+")[0]

    expected = dummy_tesseract_module.apply(test_input_val)
    expected = output_to_bytes(expected, output_format, base_dir=tmpdir)
    expected = load_bytes(expected, output_format)

    outfile = tmpdir / f"results.{container}"

    with open(outfile, "rb") as f:
        result = load_bytes(f.read(), output_format)

    def _replace_binref_paths(obj):
        # Replace binref paths with a placeholder so they can be compared
        if isinstance(obj, dict):
            if (
                obj.get("object_type") == "array"
                and obj["data"]["encoding"] == "binref"
            ):
                obj["data"]["buffer"] = "path/to/binref.bin"
                return obj
            return {k: _replace_binref_paths(v) for k, v in obj.items()}
        elif isinstance(obj, list):
            return [_replace_binref_paths(x) for x in obj]
        return obj

    expected = _replace_binref_paths(expected)
    result = _replace_binref_paths(result)

    if container == "json":
        assert result == expected
    elif container == "msgpack":
        # result contains numpy arrays, so we need to compare them separately
        assert (result["result"] == expected["result"]).all()
    else:
        # Unreachable
        raise AssertionError(f"Unexpected output format: {output_format}")


@pytest.mark.parametrize(
    "test_server",
    [
        "test_s3_server",
        "test_http_server",
    ],
)
def test_input_vals_from_network(
    cli, cli_runner, dummy_tesseract_module, test_server, request
):
    """Test the apply command with input arguments from a remote JSON file."""
    server_path = request.getfixturevalue(test_server)

    result = cli_runner.invoke(
        cli,
        ["apply", f"@{server_path}/foo.json"],
        catch_exceptions=False,
    )
    assert result.exit_code == 0, result.stderr

    test_input_val = dummy_tesseract_module.InputSchema.model_validate(test_input)
    expected = dummy_tesseract_module.apply(test_input_val).model_dump_json()
    assert json.loads(result.stdout) == json.loads(expected)


@pytest.mark.parametrize(
    "test_server",
    [
        "test_s3_server",
        "test_http_server",
    ],
)
def test_input_vals_from_network_binref(
    cli, cli_runner, dummy_tesseract_module, test_server, request
):
    """Test the apply command with input arguments from a remote JSON file."""
    server_path = request.getfixturevalue(test_server)

    # construct payload with networks paths
    test_input_network = deepcopy(test_input_binref)
    test_input_network["a"]["data"]["buffer"] = f"{server_path}/array_a.bin"
    test_input_network["b"]["data"]["buffer"] = f"{server_path}/array_b.bin"

    result = cli_runner.invoke(
        cli,
        ["apply", json.dumps({"inputs": test_input_network})],
        catch_exceptions=False,
    )
    assert result.exit_code == 0, result.stderr

    test_input_val = dummy_tesseract_module.InputSchema.model_validate(test_input)
    expected = dummy_tesseract_module.apply(test_input_val).model_dump_json()
    assert json.loads(result.stdout) == json.loads(expected)


def test_help_contains_docstring(cli, cli_runner):
    summary = "Multiplies a vector `a` by `s`, and sums the result to `b`."
    result = cli_runner.invoke(cli, ["apply", "--help"], catch_exceptions=False)
    assert result.exit_code == 0, result.stderr
    assert summary in result.stdout


def test_optional_arguments_stay_optional_in_cli(
    cli, cli_runner, dummy_tesseract_module
):
    test_input_missing = {"a": [1.0, 2.0, 3.0], "b": [1, 1, 1]}
    result = cli_runner.invoke(
        cli,
        [
            "apply",
            json.dumps({"inputs": test_input_missing}),
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 0, result.stderr

    test_input_val = dummy_tesseract_module.InputSchema.model_validate(
        test_input_missing
    )
    expected = dummy_tesseract_module.apply(test_input_val).model_dump_json()
    assert json.loads(result.stdout) == json.loads(expected)


def test_apply_fails_if_required_args_missing(cli, cli_runner):
    result = cli_runner.invoke(cli, ["apply", json.dumps({"inputs": {"a": [1, 2, 3]}})])
    print("Result:", result)
    print(result.exit_code, result.stdout, result.stderr)
    assert result.exit_code == 2
    assert "missing" in result.stderr


def test_stdout_redirect_cli():
    """Ensure that stdout is redirected to stderr during normal Python execution."""
    # Use subprocess to ensure that CLI entrypoint is used
    result = subprocess.run(
        [sys.executable, tesseract_core.runtime.cli.__file__, "--help"],
        capture_output=True,
    )
    assert result.returncode == 0, result.stderr
    assert result.stdout == b""
    assert "Usage:" in result.stderr.decode("utf-8")


def test_stdout_redirect_subprocess(tmpdir):
    """Ensure that stdout is redirected to stderr even in non-Python subprocesses."""
    # Use subprocess since pytest messes with stdout/stderr
    testscript = [
        # Print messages that signify where the output is supposed to go
        "import os",
        "import sys",
        "from tesseract_core.runtime.cli import stdout_to_stderr",
        "print('stdout', file=sys.stdout)",
        "print('stderr', file=sys.stderr)",
        "with stdout_to_stderr() as orig_stdout:",
        "    os.system('echo stderr')",
        "    print('stderr', file=sys.stdout)",
        "    print('stderr', file=sys.stderr)",
        "    print('stdout', file=orig_stdout)",
        "os.system('echo stdout')",
        "print('stdout', file=sys.stdout)",
        "print('stderr', file=sys.stderr)",
    ]
    testscript_path = tmpdir / "testscript.py"
    with open(testscript_path, "w") as f:
        f.write("\n".join(testscript))

    result = subprocess.run([sys.executable, testscript_path], capture_output=True)
    assert result.returncode == 0, (result.stdout, result.stderr)
    assert result.stdout == b"stdout\n" * 4
    assert result.stderr == b"stderr\n" * 5


def test_suggestion_on_misspelled_command(cli, cli_runner):
    result = cli_runner.invoke(cli, ["aply"], catch_exceptions=False)
    assert result.exit_code == 2, result.stdout
    assert "No such command 'aply'." in result.stderr
    assert "Did you mean 'apply'?" in result.stderr

    result = cli_runner.invoke(cli, ["applecomputersinc"], catch_exceptions=False)
    assert result.exit_code == 2, result.stdout
    assert "No such command 'applecomputersinc'." in result.stderr
    assert "Did you mean" not in result.stderr


def test_check(cli, cli_runner, dummy_tesseract_package):
    from tesseract_core.runtime.config import update_config

    tesseract_api_file = Path(dummy_tesseract_package) / "tesseract_api.py"
    update_config(api_path=tesseract_api_file)
    result = cli_runner.invoke(cli, ["check"], catch_exceptions=True)
    assert result.exit_code == 0, result.stderr
    assert "check successful" in result.stdout

    api_file_bad_syntax = Path(dummy_tesseract_package) / "tesseract_api_bad_syntax.py"
    with open(api_file_bad_syntax, "w") as f:
        f.write("bad-syntax=1")

    update_config(api_path=api_file_bad_syntax)
    result = cli_runner.invoke(cli, ["check"], catch_exceptions=True)
    assert result.exit_code == 1, result.stderr
    assert "Could not load module" in result.exception.args[0]
    full_traceback = "".join(traceback.format_exception(*result.exc_info))
    assert "SyntaxError" in full_traceback

    # Write new file for each case instead of overwriting same file to avoid caching issues
    api_file_bad_import = Path(dummy_tesseract_package) / "tesseract_api_bad_import.py"
    with open(api_file_bad_import, "w") as f:
        f.write("import non_existent_module")

    update_config(api_path=api_file_bad_import)
    result = cli_runner.invoke(cli, ["check"], catch_exceptions=True)
    assert result.exit_code == 1, result.stderr
    assert "Could not load module" in result.exception.args[0]
    full_traceback = "".join(traceback.format_exception(*result.exc_info))
    assert "ModuleNotFoundError" in full_traceback

    with open(tesseract_api_file) as f:
        tesseract_api_code = f.read()

    # check if error is raised if schemas or required endpoints are missing
    for schema_or_endpoint in ("InputSchema", "OutputSchema", "apply"):
        invalid_code = tesseract_api_code.replace(
            f"{schema_or_endpoint}", f"{schema_or_endpoint}_hide"
        )
        api_file_missing_schema_or_endpoint = (
            Path(dummy_tesseract_package)
            / f"tesseract_api_missing_{schema_or_endpoint.lower()}.py"
        )
        with open(api_file_missing_schema_or_endpoint, "w") as f:
            f.write(invalid_code)
        update_config(api_path=api_file_missing_schema_or_endpoint)
        result = cli_runner.invoke(cli, ["check"], catch_exceptions=True)
        assert result.exit_code == 1, result.stderr
        full_traceback = "".join(traceback.format_exception(*result.exc_info))
        assert (
            f"{schema_or_endpoint} is not defined in Tesseract API module"
            in result.exception.args[0]
        )

    # check if error is raised for schemas with unsupported parent class
    for schema_name in ("InputSchema", "OutputSchema"):
        invalid_code = tesseract_api_code.replace(
            f"{schema_name}(BaseModel)", f"{schema_name}(tuple)"
        )
        api_file_bad_parent_class = (
            Path(dummy_tesseract_package)
            / f"tesseract_api_bad_{schema_name.lower()}_parent_class.py"
        )
        with open(api_file_bad_parent_class, "w") as f:
            f.write(invalid_code)
        update_config(api_path=api_file_bad_parent_class)
        result = cli_runner.invoke(cli, ["check"], catch_exceptions=True)
        assert result.exit_code == 1, result.stderr
        full_traceback = "".join(traceback.format_exception(*result.exc_info))
        assert (
            f"{schema_name} is not a subclass of pydantic.BaseModel"
            in result.exception.args[0]
        )



import matplotlib.pyplot as plt
import numpy as np
from animate import make_animation
from scipy.optimize import minimize

from tesseract_core import Tesseract

# Instantiate python api to the tesseract (read port from logs above)
tesseract = Tesseract(url="http://localhost:58354")


def rosenbrock(x: np.ndarray) -> float:
    """Wrap tesseract.apply to adhere to scipy's minimize interface."""
    output = tesseract.apply({"x": x[0], "y": x[1]})
    return output["result"].item()


def rosenbrock_gradient(x: np.ndarray) -> np.ndarray:
    """Wrap tesseract.jacobian to adhere to scipy's minimize interface."""
    output = tesseract.jacobian(
        {"x": x[0], "y": x[1]}, jac_inputs=["x", "y"], jac_outputs=["result"]
    )
    return np.array([output["result"]["x"], output["result"]["y"]])


# Initial guess for the variables
x0 = np.array([-0.5, 1.0])
trajectory = [x0]

# kick off scipy's optimization using tesseract apply/jacobian
result = minimize(
    rosenbrock,
    x0,
    method="BFGS",
    jac=rosenbrock_gradient,
    options={"disp": True},
    callback=lambda xs: trajectory.append(xs.tolist()),
)

anim = make_animation(*list(zip(*trajectory)))
# anim.save("rosenbrock_optimization.gif", writer="pillow", fps=2, dpi=150)
plt.show()



"""Custom hook to inject config from runtime pyproject.toml into the main pyproject.toml.

This allows us to specify runtime dependencies and scripts in only one place.
"""

from pathlib import Path

import toml
from hatchling.metadata.plugin.interface import MetadataHookInterface

RUNTIME_PYPROJECT_PATH = "tesseract_core/runtime/meta/pyproject.toml"

BASE_OPTIONAL_DEPS = {
    "docs": [
        "sphinx",
        "sphinx_autodoc_typehints",
        "furo",
        "myst-nb",
        "sphinx_click",
        "autodoc_pydantic",
        "sphinx_design",
        "sphinx_copybutton",
        "sphinxext_opengraph",
        "tesseract-core[runtime]",
    ],
    "dev": [
        "docker",
        "fastapi",
        "httpx",  # required by fastapi test client
        "jsf",
        "mlflow",
        "numpy",
        "pre-commit",
        "pytest",
        "pytest-cov",
        "pytest-mock",
        "moto[server]",
        "aiobotocore>=2.19.0",  # without this pip dependency resolution fails
        "typeguard",
        # also add all other extras here
        "tesseract-core[runtime]",
        "tesseract-core[docs]",
    ],
    "runtime": [],
}

BASE_SCRIPTS = {
    "tesseract": "tesseract_core.sdk.cli:entrypoint",
}


class RuntimeDepenencyHook(MetadataHookInterface):
    """Injects runtime dependencies and scripts from a separate pyproject.toml file."""

    PLUGIN_NAME = "runtime-deps"

    def update(self, metadata: dict) -> dict:
        """Update the metadata with runtime dependencies and scripts."""
        runtime_metadata = toml.load(Path(self.root) / RUNTIME_PYPROJECT_PATH)
        metadata["optional-dependencies"] = {
            **BASE_OPTIONAL_DEPS,
            "runtime": runtime_metadata["project"]["dependencies"],
        }
        metadata["scripts"] = {**BASE_SCRIPTS, **runtime_metadata["project"]["scripts"]}
        return metadata



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import os
from pathlib import Path
from typing import Any

from pydantic import BaseModel, ConfigDict, FilePath


class RuntimeConfig(BaseModel):
    """Available runtime configuration."""

    api_path: FilePath = Path("tesseract_api.py")
    name: str = "Tesseract"
    description: str = ""
    version: str = "0+unknown"
    debug: bool = False
    input_path: str = "."
    output_path: str = "."

    model_config = ConfigDict(frozen=True, extra="forbid")


def update_config(**kwargs: Any) -> None:
    """Create a new runtime configuration from the current environment.

    Passed keyword arguments will override environment variables.
    """
    global _current_config

    conf_settings = {}
    for field in RuntimeConfig.model_fields.keys():
        env_key = f"TESSERACT_{field.upper()}"
        if env_key in os.environ:
            conf_settings[field] = os.environ[env_key]

    for field in _config_overrides:
        conf_settings[field] = getattr(_current_config, field)

    conf_settings.update(kwargs)
    config = RuntimeConfig(**conf_settings)

    _config_overrides.update(set(conf_settings.keys()))
    _current_config = config


_current_config = None
_config_overrides = set()


def get_config() -> RuntimeConfig:
    """Return the current runtime configuration."""
    if _current_config is None:
        update_config()
    assert _current_config is not None
    return _current_config



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

# Tesseract API module for cuda-image
# Generated by tesseract 0.8.2.dev14+g494ed54.d20250401 on 2025-04-04T18:55:51.909676


from pydantic import BaseModel

#
# Schemas
#


class InputSchema(BaseModel):
    pass


class OutputSchema(BaseModel):
    pass


#
# Required endpoints
#


def apply(inputs: InputSchema) -> OutputSchema:
    """Assert that CUDA is available."""
    from ctypes.util import find_library

    cudart = find_library("cudart")
    assert cudart is not None, "CUDA runtime library not found"
    return OutputSchema()


#
# Optional endpoints
#

# import numpy as np

# def jacobian(inputs: InputSchema, jac_inputs: set[str], jac_outputs: set[str]):
#     return {}

# def jacobian_vector_product(
#     inputs: InputSchema,
#     jvp_inputs: set[str],
#     jvp_outputs: set[str],
#     tangent_vector: dict[str, np.typing.ArrayLike]
# ) -> dict[str, np.typing.ArrayLike]:
#     return {}

# def vector_jacobian_product(
#     inputs: InputSchema,
#     vjp_inputs: set[str],
#     vjp_outputs: set[str],
#     cotangent_vector: dict[str, np.typing.ArrayLike]
# ) -> dict[str, np.typing.ArrayLike]:
#     return {}

# def abstract_eval(abstract_inputs):
#     return {}



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import logging
import logging.handlers
import sys
import warnings
from collections.abc import Iterable
from types import ModuleType
from typing import Any

import typer
from rich.console import Console
from rich.markup import escape
from rich.traceback import Traceback

DEFAULT_CONSOLE = Console(stderr=True)

LEVEL_PREFIX = {
    "DEBUG": " [dim]\\[+][/] ",
    "INFO": " [dim]\\[[/][blue]i[/][dim]][/] ",
    "WARNING": " \\[[yellow]![/]] ",
    "ERROR": " [red]\\[-][/] ",
    "CRITICAL": " [red reverse]\\[x][/] ",
}


class RichLogger(logging.Handler):
    """A logging handler that uses rich to render logs and exceptions.

    This is a pared-down version of `rich.logging.RichHandler` that only applies styling without
    additional features like word wrapping.
    """

    def __init__(
        self,
        console: Console,
        level: int | str = logging.NOTSET,
        rich_tracebacks: bool = True,
        tracebacks_suppress: Iterable[str | ModuleType] = (),
    ) -> None:
        super().__init__(level=level)
        self._console = console
        self._rich_tracebacks = rich_tracebacks
        self._tracebacks_suppress = tracebacks_suppress

    def emit(self, record: logging.LogRecord) -> None:
        """Emit a log record."""
        _exc_info = None
        if self._rich_tracebacks:
            _exc_info = record.exc_info
            # Prevent printing the traceback twice
            record.exc_info = None

        log_line = self.format(record)
        self._console.print(log_line, markup=True, soft_wrap=True)

        if self._rich_tracebacks and _exc_info:
            self._console.print(
                Traceback.from_exception(*_exc_info, suppress=self._tracebacks_suppress)
            )


def set_logger(
    level: str, catch_warnings: bool = False, rich_format: bool | None = None
) -> logging.Logger:
    """Initialize loggers."""
    level = level.upper()

    package_logger = logging.getLogger("tesseract")
    package_logger.setLevel("DEBUG")  # send everything to handlers

    if rich_format is None:
        rich_format = DEFAULT_CONSOLE.is_terminal

    if rich_format:
        ch = RichLogger(
            console=DEFAULT_CONSOLE,
            level=level,
            rich_tracebacks=True,
            tracebacks_suppress=[typer],
        )

        class PrefixFormatter(logging.Formatter):
            def format(self, record: Any, *args: Any) -> Any:
                record.levelprefix = LEVEL_PREFIX.get(record.levelname, "")
                record.msg = escape(record.msg)
                return super().format(record, *args)

        fmt = "{levelprefix!s}{message!s}"
        ch_fmt = PrefixFormatter(fmt, style="{")
    else:
        ch = logging.StreamHandler(sys.stderr)
        ch.setLevel(level)
        fmt = "{asctime} [{levelname}] {message}"
        ch_fmt = logging.Formatter(fmt, style="{")

    ch.setFormatter(ch_fmt)
    package_logger.handlers = [ch]

    if catch_warnings:
        logging.captureWarnings(True)
        warnings_logger = logging.getLogger("py.warnings")
        warnings_logger.handlers = [ch]
        warnings_logger.setLevel(level)

        def custom_formatwarning(
            message: str,
            category: Any,
            filename: str,
            lineno: int,
            line: str | None = None,
        ) -> str:
            if rich_format:
                out = f"[yellow]{category.__name__}[/]: {message}"
            else:
                out = f"{category.__name__}: {message}"
            return out

        warnings.formatwarning = custom_formatwarning

    return package_logger


def set_loglevel(level: str) -> None:
    """Update the log level of all loggers."""
    level = level.upper()
    package_logger = logging.getLogger("tesseract")
    for handler in package_logger.handlers:
        handler.setLevel(level)



from types import SimpleNamespace

import numpy as np
import pytest
from pydantic import ValidationError

from tesseract_core import Tesseract
from tesseract_core.sdk.tesseract import (
    HTTPClient,
    _decode_array,
    _encode_array,
    _tree_map,
)


@pytest.fixture
def mock_serving(mocker):
    serve_mock = mocker.patch("tesseract_core.sdk.engine.serve")
    serve_mock.return_value = "proj-id-123"

    fake_container = SimpleNamespace()
    fake_container.host_port = 1234
    fake_container.id = "container-id-123"

    get_project_containers_mock = mocker.patch(
        "tesseract_core.sdk.engine.get_project_containers"
    )
    get_project_containers_mock.return_value = [fake_container]

    teardown_mock = mocker.patch("tesseract_core.sdk.engine.teardown")
    logs_mock = mocker.patch("tesseract_core.sdk.engine.logs")
    return {
        "serve_mock": serve_mock,
        "get_project_containers_mock": get_project_containers_mock,
        "teardown_mock": teardown_mock,
        "logs_mock": logs_mock,
    }


@pytest.fixture
def mock_clients(mocker):
    mocker.patch("tesseract_core.sdk.tesseract.HTTPClient.run_tesseract")


def test_Tesseract_init():
    # Instantiate with a url
    t = Tesseract(url="localhost")

    # Using it as a context manager should be a no-op
    with t:
        pass


def test_Tesseract_from_tesseract_api(dummy_tesseract_location, dummy_tesseract_module):
    all_endpoints = [
        "apply",
        "jacobian",
        "jacobian_vector_product",
        "vector_jacobian_product",
        "health",
        "input_schema",
        "output_schema",
        "abstract_eval",
    ]

    t = Tesseract.from_tesseract_api(dummy_tesseract_location / "tesseract_api.py")
    endpoints = t.available_endpoints
    assert endpoints == all_endpoints

    # should also work when importing the module
    t = Tesseract.from_tesseract_api(dummy_tesseract_module)
    endpoints = t.available_endpoints
    assert endpoints == all_endpoints


def test_Tesseract_from_image(mock_serving, mock_clients):
    # Object is built and has the correct attributes set
    t = Tesseract.from_image(
        "sometesseract:0.2.3", input_path="/my/files", gpus=["all"]
    )

    # Now we can use it as a context manager
    # NOTE: we invoke available_endpoints because it requires an active client and is not cached
    with t:
        _ = t.available_endpoints

    # Trying to use methods from outside the context manager should raise
    with pytest.raises(RuntimeError):
        _ = t.available_endpoints

    # Works if we serve first
    try:
        t.serve()
        _ = t.available_endpoints
    finally:
        t.teardown()


def test_Tesseract_schema_methods(mocker, mock_serving):
    mocked_run = mocker.patch("tesseract_core.sdk.tesseract.HTTPClient.run_tesseract")
    mocked_run.return_value = {"#defs": {"some": "stuff"}}

    with Tesseract.from_image("sometesseract:0.2.3") as t:
        input_schema = t.input_schema
        output_schema = t.output_schema
        openapi_schema = t.openapi_schema

    assert input_schema == output_schema == openapi_schema == mocked_run.return_value


def test_serve_lifecycle(mock_serving, mock_clients):
    t = Tesseract.from_image("sometesseract:0.2.3")

    with t:
        pass

    mock_serving["serve_mock"].assert_called_with(
        ["sometesseract:0.2.3"],
        ports=None,
        volumes=[],
        environment={},
        gpus=None,
        debug=True,
        num_workers=1,
        host_ip="127.0.0.1",
        no_compose=False,
    )

    mock_serving["teardown_mock"].assert_called_with("proj-id-123")

    # check that the same Tesseract obj cannot be used to instantiate two containers
    with pytest.raises(RuntimeError):
        with t:
            with t:
                pass


def test_HTTPClient_run_tesseract(mocker):
    mock_response = mocker.Mock()
    mock_response.json.return_value = {"result": [4, 4, 4]}
    mock_response.raise_for_status = mocker.Mock()

    mocked_request = mocker.patch(
        "requests.request",
        return_value=mock_response,
    )

    client = HTTPClient("somehost")

    out = client.run_tesseract("apply", {"inputs": {"a": 1}})

    assert out == {"result": [4, 4, 4]}
    mocked_request.assert_called_with(
        method="POST",
        url="http://somehost/apply",
        json={"inputs": {"a": 1}},
    )


def test_HTTPClient_run_tesseract_raises_validation_error(mocker):
    mock_response = mocker.Mock()
    mock_response.json.return_value = {
        "detail": [
            {
                "type": "missing",
                "loc": ["body", "inputs", "a"],
                "msg": "Field required",
                "input": {"whoops": "whatever"},
            },
            {
                "type": "missing",
                "loc": ["body", "inputs", "b"],
                "msg": "Field required",
                "input": {"whoops": "whatever"},
            },
            {
                "type": "extra_forbidden",
                "loc": ["body", "inputs", "whoops"],
                "msg": "Extra inputs are not permitted",
                "input": "whatever",
            },
            {
                "type": "value_error",
                "loc": ["body", "inputs", "bar"],
                "msg": "Value error, Dimensionality mismatch: 2D array cannot be cast to 1D",
                "input": [[1.0, 2.0], [2.0, 3.0], [3.0, 4.0]],
                "error": {},
            },
        ]
    }
    mock_response.status_code = 422

    mocker.patch(
        "requests.request",
        return_value=mock_response,
    )

    client = HTTPClient("somehost")

    with pytest.raises(ValidationError) as excinfo:
        client.run_tesseract("apply", {"inputs": {"whoops": "whatever"}})

    # This checks as well that no duplicate "Value error" is in msg
    assert (
        excinfo.value.errors()[3]["msg"]
        == "Value error, Dimensionality mismatch: 2D array cannot be cast to 1D"
    )


@pytest.mark.parametrize(
    "b64, expected_data",
    [
        (True, {"buffer": "AACAPwAAAEAAAEBA", "encoding": "base64"}),
        (False, {"buffer": [1.0, 2.0, 3.0], "encoding": "raw"}),
    ],
)
def test_encode_array(b64, expected_data):
    a = np.array([1.0, 2.0, 3.0], dtype="float32")

    encoded = _encode_array(a, b64=b64)

    assert encoded["shape"] == (3,)
    assert encoded["dtype"] == "float32"
    assert encoded["data"] == expected_data


@pytest.mark.parametrize(
    "encoded, expected",
    [
        (
            {
                "shape": (3,),
                "dtype": "float32",
                "data": {"buffer": [1.0, 2.0, 3.0], "encoding": "raw"},
            },
            np.array([1.0, 2.0, 3.0], dtype="float32"),
        ),
        (
            {
                "shape": (1, 3),
                "dtype": "float32",
                "data": {"buffer": "AACAPwAAAEAAAEBA", "encoding": "base64"},
            },
            np.array([[1.0, 2.0, 3.0]], dtype="float32"),
        ),
    ],
)
def test_decode_array(encoded, expected):
    decoded = _decode_array(encoded)
    np.testing.assert_array_equal(decoded, expected, strict=True)


def test_tree_map():
    tree = {
        "a": [10, 20],
        "b": {"c": np.array([30])},
        "d": {"e": np.array([1.0, 2.0, 3.0])},
        "f": "hello",
    }

    encoded = _tree_map(_encode_array, tree, is_leaf=lambda x: hasattr(x, "shape"))

    assert encoded == {
        "a": [10, 20],
        "b": {
            "c": {
                "shape": (1,),
                "dtype": "int64",
                "data": {"buffer": "HgAAAAAAAAA=", "encoding": "base64"},
            }
        },
        "d": {
            "e": {
                "shape": (3,),
                "dtype": "float64",
                "data": {
                    "buffer": "AAAAAAAA8D8AAAAAAAAAQAAAAAAAAAhA",
                    "encoding": "base64",
                },
            }
        },
        "f": "hello",
    }



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import re
from collections.abc import Sequence
from pathlib import Path
from typing import Annotated, Any, Literal, Optional, Union, get_args
from uuid import uuid4

import numpy as np
import pybase64
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    JsonValue,
    PositiveInt,
    ValidationInfo,
    create_model,
)

from tesseract_core.runtime.file_interactions import (
    get_filesize,
    is_absolute_path,
    is_url,
    join_paths,
    read_from_path,
    write_to_path,
)

AllowedDtypes = Literal[
    "float16",
    "float32",
    "float64",
    "int8",
    "int16",
    "int32",
    "int64",
    "bool",
    "uint8",
    "uint16",
    "uint32",
    "uint64",
    "complex64",
    "complex128",
]
EllipsisType = type(Ellipsis)
ArrayLike = Union[np.ndarray, np.number, np.bool_]
ShapeType = Union[tuple[Optional[int], ...], EllipsisType]

MAX_BINREF_BUFFER_SIZE = 100 * 1024 * 1024  # 100 MB


# Base classes for the different array encodings
# The actual models are created dynamically based on the expected shape and dtype by get_array_model


class Base64ArrayData(BaseModel):
    """Data structure for base64 encoded binary buffers."""

    buffer: Annotated[
        str,
        Field(
            description="Base64 encoded binary buffer",
            examples=["<base64 encoded string>"],
        ),
    ]
    encoding: Literal["base64"]
    model_config = ConfigDict(extra="forbid")


class BinrefArrayData(BaseModel):
    """Data structure that dumps array data to binary file."""

    buffer: str = Field(pattern=r"^.+?(\:\d+)?$")
    encoding: Literal["binref"]
    model_config = ConfigDict(extra="forbid")


class JsonArrayData(BaseModel):
    """Data structure for json buffers (list of decimal numbers)."""

    buffer: JsonValue
    encoding: Literal["json"]
    model_config = ConfigDict(extra="forbid")


class EncodedArrayModel(BaseModel):
    """Base class for general encoded arrays.

    Allowed values for shape and dtype are enforced by subclasses.
    """

    object_type: Literal["array"]
    shape: tuple[PositiveInt, ...]
    dtype: AllowedDtypes
    data: Union[BinrefArrayData, Base64ArrayData, JsonArrayData]
    model_config = ConfigDict(extra="forbid")


def get_array_model(
    expected_shape: ShapeType, expected_dtype: Optional[str], flags: Sequence[str]
) -> type[EncodedArrayModel]:
    """Create a Pydantic model for an encoded array that does validation on the given expected shape and dtype."""
    if expected_dtype is None:
        dtype_type = AllowedDtypes
    else:
        # Only allow dtypes that can be cast to the expected dtype
        subdtypes = [
            dtype
            for dtype in get_args(AllowedDtypes)
            if np.can_cast(dtype, expected_dtype, casting="same_kind")
        ]
        dtype_type = Literal[tuple(subdtypes)]

    shape_kwargs = {}

    # Only allow shapes that can be broadcasted to the expected shape
    if expected_shape is Ellipsis:
        # No shape check
        shape_type = tuple[int, ...]
    else:
        # There are 3 cases for each dimension `n`:
        # - n=None: polymorphic dimension, can be any positive int
        # - n=1: fixed dimension, must be 1
        # - n=N: fixed dimension, must be N or 1 (triggers broadcasting to N)
        # Example: expected_shape=(None, 1, 3) -> allowed_vals=tuple[PositiveInt, Literal[1], Literal[1, 3]]
        allowed_vals = []
        for dim in expected_shape:
            if dim is None:
                allowed_vals.append(PositiveInt)
            elif dim == 1:
                allowed_vals.append(Literal[1])
            else:
                allowed_vals.append(Literal[1, dim])

        if not allowed_vals:
            # Scalar -> require empty tuple
            shape_type = tuple[int, ...]

        shape_type = tuple[tuple(allowed_vals)]
        shape_kwargs.update(
            examples=([1 if s is None else s for s in expected_shape],),
            # Dimensionality must match exactly
            min_length=len(expected_shape),
            max_length=len(expected_shape),
            # TODO: This is a hack to allow JSF to parse the JSON schema
            # see https://github.com/ghandic/jsf/issues/118
            json_schema_extra={"items": {"type": "integer"}},
        )

    # Add flags to the model config
    config = EncodedArrayModel.model_config
    config["json_schema_extra"] = {"array_flags": flags}

    fields = {
        "object_type": (
            Literal["array"],
            Field(
                description="Indicates that this dict can be parsed to an array.",
                default="array",
            ),
        ),
        "shape": (
            shape_type,
            Field(
                description="Shape of the array",
                **shape_kwargs,
            ),
        ),
        "dtype": (
            dtype_type,
            Field(
                description="Data type of the array",
                examples=[expected_dtype or "float64"],
            ),
        ),
        # Choose the appropriate data structure based on the encoding
        "data": (
            Union[BinrefArrayData, Base64ArrayData, JsonArrayData],
            Field(discriminator="encoding"),
        ),
        "model_config": (ConfigDict, config),
    }

    if expected_shape is Ellipsis:
        readable_shape = "anyrank"
    elif not expected_shape:
        readable_shape = "scalar"
    else:
        readable_shape = "_".join(
            str(s) if s is not None else "any" for s in expected_shape
        )

    readable_flags = "_".join(flags) if flags else "noflags"

    out = create_model(
        f"EncodedArrayModel__{readable_shape}__{expected_dtype}__{readable_flags}",
        **fields,
        __base__=EncodedArrayModel,
    )
    return out


def _dump_binref_arraydict(
    arr: Union[np.ndarray, np.number, np.bool_],
    base_dir: Union[Path, str],
    current_binref_uuid: str,
    max_file_size: int = MAX_BINREF_BUFFER_SIZE,
) -> tuple[dict[str, Union[str, dict[str, str]]], str]:
    """Dump array to json+binref encoded array dict.

    Writes a .bin file and returns json encoded data.
    """
    target_name = f"{current_binref_uuid}.bin"
    target_path = join_paths(base_dir, target_name)

    current_size = get_filesize(target_path)

    # if the current buffer is too large, use a new one
    if current_size > max_file_size:
        current_size = 0
        current_binref_uuid = str(uuid4())
        target_name = f"{current_binref_uuid}.bin"
        target_path = join_paths(base_dir, target_name)

    write_to_path(arr.tobytes(), target_path, append=True)
    offset = current_size

    data = {"buffer": f"{target_name}:{offset}", "encoding": "binref"}
    arraydict = {
        "object_type": "array",
        "shape": arr.shape,
        "dtype": arr.dtype.name,
        "data": data,
    }
    return arraydict, current_binref_uuid


def _dump_base64_arraydict(
    arr: Union[np.ndarray, np.number, np.bool_],
) -> dict[str, Union[str, dict[str, str]]]:
    """Dump array to json+base64 encoded array dict."""
    data = {
        "buffer": pybase64.b64encode(arr.tobytes()).decode(),
        "encoding": "base64",
    }
    arraydict = {
        "object_type": "array",
        "shape": arr.shape,
        "dtype": arr.dtype.name,
        "data": data,
    }
    return arraydict


def _dump_json_arraydict(
    arr: Union[np.ndarray, np.number, np.bool_],
) -> dict[str, Union[str, dict[str, str]]]:
    """Dump array to json encoded array dict."""
    data = {
        "buffer": arr.tolist(),
        "encoding": "json",
    }
    arraydict = {
        "object_type": "array",
        "shape": arr.shape,
        "dtype": arr.dtype.name,
        "data": data,
    }
    return arraydict


def _load_base64_arraydict(val: dict) -> np.ndarray:
    """Load array from json+base64 encoded array dict."""
    buffer = pybase64.b64decode(val["data"]["buffer"], validate=True)
    return np.frombuffer(buffer, dtype=val["dtype"]).reshape(val["shape"])


def _load_binref_arraydict(val: dict, base_dir: Union[str, Path, None]) -> np.ndarray:
    """Load array from json+binref encoded array dict."""
    path_match = re.match(r"^(?P<path>.+?)(\:(?P<offset>\d+))?$", val["data"]["buffer"])
    bufferpath = path_match.group("path")
    if path_match.group("offset") is None:
        offset = 0
    else:
        offset = int(path_match.group("offset"))

    uses_relative_path = not is_absolute_path(bufferpath) and not is_url(bufferpath)
    if uses_relative_path and base_dir is None:
        raise ValueError(
            "Array data is binref encoded with a relative path but no base_dir is provided. "
            "Invoke the Tesseract via file aliasing or make sure that paths are absolute."
        )

    dtype = np.dtype(val["dtype"])
    shape = val["shape"]
    size = 1 if len(shape) == 0 else np.prod(shape)
    num_bytes = int(size * dtype.itemsize)

    if base_dir is not None:
        bufferpath = join_paths(base_dir, bufferpath)

    buffer = read_from_path(bufferpath, offset=offset, length=num_bytes)
    return np.frombuffer(buffer, dtype=dtype).reshape(shape)


def _coerce_shape_dtype(
    arr: ArrayLike, expected_shape: ShapeType, expected_dtype: Optional[str]
) -> ArrayLike:
    """Coerce the shape and dtype of the passed array to the expected values."""
    if expected_shape is Ellipsis:
        # No shape check
        out_shape = arr.shape
    else:
        if len(arr.shape) != len(expected_shape):
            raise ValueError(
                f"Dimensionality mismatch: {len(arr.shape)}D array cannot be cast to {len(expected_shape)}D"
            )

        out_shape = tuple(
            # Polymorphic dims -> keep the passed shape
            arr.shape[i] if expected_shape[i] is None else expected_shape[i]
            for i in range(len(expected_shape))
        )

    # Broadcast the arr to the expected shape and dtype
    try:
        arr = np.broadcast_to(arr, out_shape)
    except ValueError:
        raise ValueError(
            f"Shape mismatch: {arr.shape} cannot be cast to {out_shape}"
        ) from None

    if expected_dtype is not None:
        if not np.can_cast(arr.dtype, expected_dtype, casting="same_kind"):
            raise ValueError(
                f"Dtype mismatch: {arr.dtype} cannot be cast to {expected_dtype}"
            )
        arr = arr.astype(expected_dtype)

    allowed_dtypes = [dtype.lower() for dtype in get_args(AllowedDtypes)]
    if arr.dtype.name not in allowed_dtypes:
        raise ValueError(
            f"Got invalid dtype: {arr.dtype.name}, expected one of {allowed_dtypes}"
        )

    if not out_shape:
        # Cast to a scalar type
        return arr.dtype.type(arr)

    return arr


def python_to_array(
    val: Any, expected_shape: ShapeType, expected_dtype: Optional[str]
) -> ArrayLike:
    """Convert a Python object to a NumPy array."""
    val = np.asarray(val, order="C")
    if not np.issubdtype(val.dtype, np.number) and not np.issubdtype(
        val.dtype, np.bool_
    ):
        raise ValueError(
            f"Could not convert object to numeric NumPy array (got dtype: {val.dtype})"
        )
    return _coerce_shape_dtype(val, expected_shape, expected_dtype)


def decode_array(
    val: EncodedArrayModel,
    info: ValidationInfo,
    expected_shape: ShapeType,
    expected_dtype: Optional[str],
) -> ArrayLike:
    """Decode an EncodedArrayModel to a NumPy array."""
    context = info.context if info.context else {}

    try:
        if val.data.encoding == "base64":
            data = _load_base64_arraydict(val.model_dump())

        elif val.data.encoding == "binref":
            data = _load_binref_arraydict(val.model_dump(), context.get("base_dir"))

        # keep checking for "raw" for backwards compat
        elif val.data.encoding in {"json", "raw"}:
            data = np.asarray(val.data.buffer).reshape(val.shape)
            if np.issubdtype(data.dtype, np.floating) and np.issubdtype(
                val.dtype, np.integer
            ):
                if np.any(data % 1):
                    raise ValueError(
                        f"Expected integer data, but got floating point data: {data}"
                    )
            data = data.astype(val.dtype, casting="unsafe")

        else:
            # Unreachable
            raise AssertionError(f"Unsupported encoding: {val.data.encoding}")

    except Exception as e:
        raise ValueError(f"Failed to decode buffer as {val.data.encoding}: {e}") from e

    data = _coerce_shape_dtype(data, expected_shape, expected_dtype)
    return data


def encode_array(
    arr: ArrayLike, info: Any, expected_shape: ShapeType, expected_dtype: Optional[str]
) -> Union[EncodedArrayModel, ArrayLike]:
    """Encode a NumPy array as an EncodedArrayModel."""
    # Convert to a NumPy array if necessary
    arr = python_to_array(arr, expected_shape, expected_dtype)

    context = info.context if info.context else {}

    # Python mode -> just return the array without encoding
    if not info.mode_is_json():
        return arr

    array_encoding = context.get("array_encoding", "json")
    if array_encoding == "base64":
        data = _dump_base64_arraydict(arr)
    elif array_encoding == "binref":
        if not context.get("base_dir"):
            raise ValueError(
                "To write data with binref encoding you have to provide a 'base_dir' to "
                "store array data like so: "
                "`context={'array_encoding':'binref', 'base_dir': 'path/to/arraydata'}"
            )
        data, new_binref_uuid = _dump_binref_arraydict(
            arr,
            base_dir=context.get("base_dir"),
            current_binref_uuid=context.get("__binref_uuid", str(uuid4())),
            max_file_size=context.get("max_file_size", MAX_BINREF_BUFFER_SIZE),
        )
        context["__binref_uuid"] = new_binref_uuid
    elif array_encoding == "json":
        data = _dump_json_arraydict(arr)
    else:
        # Unreachable
        raise AssertionError(f"Unsupported encoding: {array_encoding}")

    return EncodedArrayModel.model_validate(data)



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""Docker client for Tesseract usage."""

import json
import logging
import os
import shlex
import subprocess
from dataclasses import dataclass
from pathlib import Path
from textwrap import indent
from typing import Literal

from tesseract_core.sdk.config import get_config

logger = logging.getLogger("tesseract")


# store a reference to list, which is shadowed by some function names below
list_ = list


def _get_executable(program: Literal["docker", "docker-compose"]) -> tuple[str, ...]:
    config = get_config()
    if program == "docker":
        return config.docker_executable
    if program == "docker-compose":
        return config.docker_compose_executable
    raise ValueError(f"Unknown program: {program}")


def is_podman() -> bool:
    """Check if the current environment is using Podman instead of Docker."""
    docker = _get_executable("docker")
    try:
        result = subprocess.run(
            [*docker, "version"],
            capture_output=True,
            text=True,
            check=True,
        )
        return "podman" in result.stdout.lower()
    except subprocess.CalledProcessError:
        return False


@dataclass
class Image:
    """Image class to wrap Docker image details."""

    id: str | None
    short_id: str | None
    tags: list[str] | None
    attrs: dict

    @classmethod
    def from_dict(cls, json_dict: dict) -> "Image":
        """Create an Image object from a json dictionary."""
        image_id = json_dict.get("Id", None)
        short_id = None
        if image_id:
            if image_id.startswith("sha256:"):
                short_id = image_id[:19]
            else:
                # Some container engines (e.g., Podman) do not prefix IDs with sha256
                short_id = image_id[:12]

        return cls(
            id=image_id,
            short_id=short_id,
            tags=json_dict.get("RepoTags", None),
            attrs=json_dict,
        )


class Images:
    """Namespace for functions to interface with Tesseract docker images."""

    @staticmethod
    def get(image_id_or_name: str | bytes, tesseract_only: bool = True) -> Image:
        """Returns the metadata for a specific image.

        In docker-py, there is no substring matching and the image name is the
        last tag in the list of tags, so if an image has multiple tags, only
        one of the tags would be able to find the image.

        However, in podman, this is not the case. Podman has substring matching
        by "/" segments to handle repository urls and returns images even if
        partial name is specified, or if image has multiple tags.

        We chose to support podman's largest string matching functionality here.

        Params:
            image_id_or_name: The image name or id to get.
            tesseract_only: If True, only retrieves Tesseract images.

        Returns:
            Image object.
        """
        if not image_id_or_name:
            raise ValueError("Image name cannot be empty.")

        docker = _get_executable("docker")
        try:
            result = subprocess.run(
                [*docker, "inspect", image_id_or_name, "--type", "image"],
                check=True,
                capture_output=True,
                text=True,
            )
            json_dict = json.loads(result.stdout)
        except subprocess.CalledProcessError as ex:
            raise ImageNotFound(f"Image {image_id_or_name} not found.") from ex
        if not json_dict:
            raise ImageNotFound(f"Image {image_id_or_name} not found.")

        if tesseract_only and not any(
            "TESSERACT_NAME" in env_var for env_var in json_dict[0]["Config"]["Env"]
        ):
            raise ImageNotFound(f"Image {image_id_or_name} is not a Tesseract image.")
        image_obj = Image.from_dict(json_dict[0])
        return image_obj

    @staticmethod
    def list(tesseract_only: bool = True) -> list_[Image]:
        """Returns the current list of images.

        Params:
            tesseract_only: If True, only return Tesseract images.

        Returns:
            List of Image objects.
        """
        return Images._get_images(tesseract_only=tesseract_only)

    @staticmethod
    def remove(image: str) -> None:
        """Remove an image (name or id) from the local Docker registry.

        Params:
            image: The image name or id to remove.
        """
        docker = _get_executable("docker")
        try:
            res = subprocess.run(
                [*docker, "rmi", image, "--force"],
                check=True,
                capture_output=True,
                text=True,
            )
        except subprocess.CalledProcessError as ex:
            raise ImageNotFound(f"Cannot remove image {image}: {ex}") from ex

        if "No such image" in res.stderr:
            raise ImageNotFound(f"Cannot remove image {image}: {res.stderr}")

    @staticmethod
    def _get_buildx_command(
        path: str | Path,
        tag: str,
        dockerfile: str | Path,
        ssh: str | None = None,
    ) -> list_[str]:
        """Get the buildx command for building Docker images.

        Returns:
            The buildx command as a list of strings.
        """
        docker = _get_executable("docker")
        extra_args = get_config().docker_build_args

        if ssh is not None:
            extra_args = ("--ssh", ssh, *extra_args)

        build_cmd = [
            *docker,
            "buildx",
            "build",
            "--load",
            "--tag",
            tag,
            "--file",
            str(dockerfile),
            *extra_args,
            "--",
            str(path),
        ]

        return build_cmd

    @staticmethod
    def buildx(
        path: str | Path,
        tag: str,
        dockerfile: str | Path,
        ssh: str | None = None,
    ) -> Image:
        """Build a Docker image from a Dockerfile using BuildKit.

        Params:
            path: Path to the directory containing the Dockerfile.
            tag: The name of the image to build.
            dockerfile: path within the build context to the Dockerfile.
            ssh: If not None, pass given argument to buildx --ssh command.

        Returns:
            Built Image object.
        """
        from tesseract_core.sdk.engine import LogPipe

        build_cmd = Images._get_buildx_command(
            path=path,
            tag=tag,
            dockerfile=dockerfile,
            ssh=ssh,
        )

        out_pipe = LogPipe(logging.DEBUG)

        with out_pipe as out_pipe_fd:
            proc = subprocess.run(build_cmd, stdout=out_pipe_fd, stderr=out_pipe_fd)

        logs = out_pipe.captured_lines
        return_code = proc.returncode

        if return_code != 0:
            raise BuildError(logs)

        return Images.get(tag)

    @staticmethod
    def _get_images(tesseract_only: bool = True) -> list_[Image]:
        """Gets the list of images by querying Docker CLI.

        Params:
            tesseract_only: If True, only return Tesseract images.

        Returns:
            List of (non-dangling) Image objects.
        """
        docker = _get_executable("docker")
        images = []
        try:
            image_ids = subprocess.run(
                [*docker, "images", "-q"],  # List only image IDs
                capture_output=True,
                text=True,
                check=True,
            )
        except subprocess.CalledProcessError as ex:
            raise APIError(f"Cannot list Docker images: {ex}") from ex

        if not image_ids.stdout:
            return []

        image_ids = image_ids.stdout.strip().split("\n")
        # Filter list to exclude empty strings.
        image_ids = [image_id for image_id in image_ids if image_id]

        # If image shows up multiple times, that means it is tagged multiple times
        # So we need to make multiple copies of the image with different names
        image_counts = {}
        for image_id in image_ids:
            image_counts[image_id] = image_counts.get(image_id, 0) + 1

        json_dicts = get_docker_metadata(
            image_ids, is_image=True, tesseract_only=tesseract_only
        )
        for _, json_dict in json_dicts.items():
            image = Image.from_dict(json_dict)
            images.append(image)

        return images


@dataclass
class Container:
    """Container class to wrap Docker container details.

    Container class has additional member variables `host_port` and `host_ip` that
    docker-py does not have. This is because Tesseract requires frequent access to the host
    port mappings.
    """

    id: str
    short_id: str
    name: str
    attrs: dict

    @classmethod
    def from_dict(cls, json_dict: dict) -> "Container":
        """Create a Container object from a json dictionary."""
        return cls(
            id=json_dict.get("Id", None),
            short_id=json_dict.get("Id", [])[:12],
            name=json_dict.get("Name", None).lstrip("/"),
            attrs=json_dict,
        )

    @property
    def image(self) -> Image | None:
        """Gets the image ID of the container."""
        image_id = self.attrs.get("ImageID", self.attrs["Image"])
        if image_id is None:
            return None
        return Images.get(image_id)

    @property
    def host_port(self) -> str | None:
        """Gets the host port of the container."""
        if self.attrs.get("NetworkSettings", None):
            ports = self.attrs["NetworkSettings"].get("Ports", None)
            if ports:
                api_port_key = "8000/tcp"
                if ports[api_port_key]:
                    return ports[api_port_key][0].get("HostPort")
        return None

    @property
    def host_debugpy_port(self) -> str | None:
        """Gets the host port which maps to debugpy server in the container."""
        if self.attrs.get("NetworkSettings", None):
            ports = self.attrs["NetworkSettings"].get("Ports", None)
            if ports:
                debugpy_port_key = "5678/tcp"
                if debugpy_port_key in ports:
                    return ports[debugpy_port_key][0].get("HostPort")
        return None

    @property
    def host_ip(self) -> str | None:
        """Gets the host IP of the container."""
        if self.attrs.get("NetworkSettings", None):
            ports = self.attrs["NetworkSettings"].get("Ports", None)
            if ports:
                api_port_key = "8000/tcp"
                if ports[api_port_key]:
                    return ports[api_port_key][0].get("HostIp")
        return None

    @property
    def project_id(self) -> str | None:
        """Gets the project ID of the container."""
        project_id = self.attrs.get("Config", None)
        if project_id:
            project_id = project_id["Labels"].get("com.docker.compose.project", None)
        return project_id

    def exec_run(self, command: list) -> tuple[int, bytes]:
        """Run a command in this container.

        Return exit code and stdout.
        """
        docker = _get_executable("docker")
        result = subprocess.run(
            [*docker, "exec", self.id, *command],
            check=False,
            capture_output=True,
            text=False,
        )
        if result.returncode != 0:
            raise ContainerError(
                self.id,
                result.returncode,
                shlex.join(command),
                self.image.id if self.image else "unknown",
                result.stderr,
            )
        return result.returncode, result.stdout

    def logs(self, stdout: bool = True, stderr: bool = True) -> bytes:
        """Get the logs for this container.

        Logs needs to be called if container is running in a detached state,
        and we wish to retrieve the logs from the command executing in the container.

        Params:
            stdout: If True, return stdout.
            stderr: If True, return stderr.
        """
        docker = _get_executable("docker")

        if stdout and stderr:
            # use subprocess.STDOUT to combine stdout and stderr into one stream
            # with the correct order of output
            stdout_pipe = subprocess.PIPE
            stderr_pipe = subprocess.STDOUT
            output_attr = "stdout"
        elif not stdout and stderr:
            stdout_pipe = subprocess.DEVNULL
            stderr_pipe = subprocess.PIPE
            output_attr = "stderr"
        elif stdout and not stderr:
            stdout_pipe = subprocess.PIPE
            stderr_pipe = subprocess.DEVNULL
            output_attr = "stdout"
        else:
            raise ValueError("At least one of stdout or stderr must be True.")

        try:
            result = subprocess.run(
                [*docker, "logs", self.id],
                check=True,
                stdout=stdout_pipe,
                stderr=stderr_pipe,
            )
        except subprocess.CalledProcessError as ex:
            raise APIError(f"Cannot get logs for container {self.id}: {ex}") from ex

        return getattr(result, output_attr)

    def wait(self) -> dict:
        """Wait for container to finish running.

        Returns:
            A dict with the exit code of the container.
        """
        docker = _get_executable("docker")

        try:
            result = subprocess.run(
                [*docker, "wait", self.id],
                check=True,
                capture_output=True,
                text=True,
            )
            # Container's exit code is printed by the wait command
            return {"StatusCode": int(result.stdout)}
        except subprocess.CalledProcessError as ex:
            raise APIError(f"Cannot wait for container {self.id}: {ex}") from ex

    def remove(self, v: bool = False, link: bool = False, force: bool = False) -> str:
        """Remove the container.

        Params:
            v: If True, remove volumes associated with the container.
            link: If True, remove links to the container.
            force: If True, force remove the container.

        Returns:
            The output of the remove command.
        """
        docker = _get_executable("docker")
        try:
            result = subprocess.run(
                [
                    *docker,
                    "rm",
                    *(["-f"] if force else []),
                    *(["-v"] if v else []),
                    *(["-l"] if link else []),
                    self.id,
                ],
                check=True,
                capture_output=True,
                text=True,
            )
            return result.stdout
        except subprocess.CalledProcessError as ex:
            if "docker" in ex.stderr:
                raise APIError(f"Cannot remove container {self.id}: {ex}") from ex
            raise ex


class Containers:
    """Namespace to interface with docker containers."""

    @staticmethod
    def list(all: bool = False, tesseract_only: bool = True) -> list_[Container]:
        """Returns the current list of containers.

        Params:
            all: If True, include stopped containers.
            tesseract_only: If True, only return Tesseract containers.

        Returns:
            List of Container objects.
        """
        return Containers._get_containers(
            include_stopped=all, tesseract_only=tesseract_only
        )

    @staticmethod
    def get(id_or_name: str, tesseract_only: bool = True) -> Container:
        """Returns the metadata for a specific container.

        Params:
            id_or_name: The container name or id to get.
            tesseract_only: If True, only retrieves Tesseract containers.

        Returns:
            Container object.
        """
        docker = _get_executable("docker")

        try:
            result = subprocess.run(
                [*docker, "inspect", id_or_name, "--type", "container"],
                check=True,
                capture_output=True,
                text=True,
            )
            json_dict = json.loads(result.stdout)
        except subprocess.CalledProcessError as ex:
            raise NotFound(f"Container {id_or_name} not found.") from ex

        if not json_dict:
            raise NotFound(f"Container {id_or_name} not found.")
        if tesseract_only and not any(
            "TESSERACT_NAME" in env_var for env_var in json_dict[0]["Config"]["Env"]
        ):
            raise NotFound(f"Container {id_or_name} is not a Tesseract container.")

        container_obj = Container.from_dict(json_dict[0])
        return container_obj

    @staticmethod
    def run(
        image: str,
        command: list_[str],
        volumes: dict | None = None,
        device_requests: list_[int | str] | None = None,
        environment: dict[str, str] | None = None,
        network: str | None = None,
        detach: bool = False,
        remove: bool = False,
        ports: dict | None = None,
        stdout: bool = True,
        stderr: bool = False,
        user: str | None = None,
        extra_args: list_[str] | None = None,
    ) -> Container | tuple[bytes, bytes] | bytes:
        """Run a command in a container from an image.

        Params:
            image: The image name or id to run the command in.
            command: The command to run in the container.
            volumes: A dict of volumes to mount in the container.
            user: String of user information to run command as in the format "uid:(optional)gid".
            device_requests: A list of device requests for the container.
            detach: If True, run the container in detached mode. Detach must be set to
                    True if we wish to retrieve the container id of the running container,
                    and if detach is true, we must wait on the container to finish
                    running and retrieve the logs of the container manually.
            remove: If remove is set to True, the container will automatically remove itself
                    after it finishes executing the command. This means that we cannot set
                    both detach and remove simulataneously to True or else there
                    would be no way of retrieving the logs from the removed container.
            ports: A dict of ports to expose in the container. The keys are the host ports
                   and the values are the container ports.
            stdout: If True, return stdout.
            stderr: If True, return stderr.
            environment: Environment variables to set in the container.
            extra_args: Additional arguments to pass to the `docker run` CLI command.

        Returns:
            Container object if detach is True, otherwise returns list of stdout and stderr.
        """
        config = get_config()
        docker = _get_executable("docker")

        if isinstance(command, str):
            command = [command]

        optional_args = []

        # Convert the parsed_volumes into a list of strings in proper argument format,
        # `-v host_path:container_path:mode`.
        if volumes:
            volume_args = []
            for host_path, volume_info in volumes.items():
                volume_args.append("-v")
                volume_args.append(
                    f"{host_path}:{volume_info['bind']}:{volume_info['mode']}"
                )
            optional_args.extend(volume_args)

        if network:
            optional_args.extend(["--network", network])

        if user:
            optional_args.extend(["-u", user])

        if device_requests:
            gpus_str = ",".join(device_requests)
            optional_args.extend(["--gpus", f'"device={gpus_str}"'])

        if environment:
            env_args = []
            for env_var, value in environment.items():
                env_args.extend(["-e", f"{env_var}={value}"])
            optional_args.extend(env_args)

        # Remove and detached cannot both be set to true
        if remove and detach:
            raise ValueError(
                "Cannot set both remove and detach to True when running a container."
            )
        if detach:
            optional_args.append("--detach")
        if remove:
            optional_args.append("--rm")

        if ports:
            for host_port, container_port in ports.items():
                optional_args.extend(["-p", f"{host_port}:{container_port}"])

        if extra_args is None:
            extra_args = []

        full_cmd = [
            *docker,
            "run",
            *optional_args,
            *config.docker_run_args,
            *extra_args,
            image,
            *command,
        ]

        logger.debug(f"Running command: {full_cmd}")

        result = subprocess.run(
            full_cmd,
            capture_output=True,
            text=False,
            check=False,
        )

        if result.returncode != 0:
            stderr_str = result.stderr.decode("utf-8", errors="ignore")
            if "repository" in stderr_str:
                raise ImageNotFound(stderr_str)
            raise ContainerError(
                None,
                result.returncode,
                shlex.join(full_cmd),
                image,
                result.stderr,
            )

        if detach:
            # If detach is True, stdout prints out the container ID of the running container
            container_id = result.stdout.decode("utf-8", errors="ignore").strip()
            container_obj = Containers.get(container_id)
            return container_obj

        if stdout and stderr:
            return result.stdout, result.stderr
        if stderr:
            return result.stderr
        return result.stdout

    @staticmethod
    def _get_containers(
        include_stopped: bool = False, tesseract_only: bool = True
    ) -> list_[Container]:
        """Updates and retrieves the list of containers by querying Docker CLI.

        Params:
            include_stopped: If True, include stopped containers.
            tesseract_only: If True, only return Tesseract containers.

        Returns:
            List of Container objects.
        """
        docker = _get_executable("docker")
        containers = []

        cmd = [*docker, "ps", "-q"]
        if include_stopped:
            cmd.append("--all")

        try:
            result = subprocess.run(
                cmd,
                capture_output=True,
                text=True,
                check=True,
            )
        except subprocess.CalledProcessError as ex:
            raise APIError(f"Cannot list Docker containers: {ex}") from ex

        if not result.stdout:
            return []

        container_ids = result.stdout.strip().split("\n")

        # Filter list to  exclude empty strings.
        container_ids = [container_id for container_id in container_ids if container_id]
        json_dicts = get_docker_metadata(container_ids, tesseract_only=tesseract_only)
        for _, json_dict in json_dicts.items():
            container = Container.from_dict(json_dict)
            containers.append(container)

        return containers


class Compose:
    """Custom namespace to interface with docker compose projects.

    There is no equivalent for this class in docker-py; however, we frequently
    interact with docker compose projects in Tesseract and this namespace makes
    such interactions easier.
    """

    @staticmethod
    def list(include_stopped: bool = False) -> dict[str, list_[str]]:
        """Returns the current list of projects.

        Params:
            include_stopped: If True, include stopped projects.

        Returns:
            Dict of projects, with the project name as the key and a list of container ids as the value.
        """
        return Compose._update_projects(include_stopped)

    @staticmethod
    def up(compose_fpath: str, project_name: str) -> str:
        """Start containers using Docker Compose template.

        Params:
            compose_fpath: Path to the Docker Compose template.
            project_name: Name of the project.

        Returns:
            The project name.
        """
        docker_compose = _get_executable("docker-compose")
        logger.info("Waiting for Tesseract containers to start ...")
        try:
            _ = subprocess.run(
                [
                    *docker_compose,
                    "-f",
                    compose_fpath,
                    "-p",
                    project_name,
                    "up",
                    "-d",
                    "--wait",
                ],
                check=True,
                capture_output=True,
            )
            return project_name
        except subprocess.CalledProcessError as ex:
            logger.error(str(ex))
            logger.error(ex.stderr.decode())
            # If the project successfully started, try to get the logs from the containers
            project_containers = Compose.list(include_stopped=True).get(
                project_name, ()
            )
            for container_name in project_containers:
                container = Containers.get(container_name)
                logger.error(f"Container {container_name} logs:")
                logger.error(indent(container.logs(stderr=True).decode(), " > "))
            raise APIError("Failed to start Tesseract containers.") from ex

    @staticmethod
    def down(project_id: str) -> bool:
        """Stop and remove containers and networks associated to a project.

        Params:
            project_id: The project name to stop.

        Returns:
            True if the project was stopped successfully, False otherwise.
        """
        docker_compose = _get_executable("docker-compose")
        try:
            __ = subprocess.run(
                [*docker_compose, "-p", project_id, "down"],
                check=True,
                capture_output=True,
            )
            return True
        except subprocess.CalledProcessError as ex:
            logger.error(str(ex))
            return False

    @staticmethod
    def exists(project_id: str) -> bool:
        """Check if Docker Compose project exists.

        Params:
            project_id: The project name to check.

        Returns:
            True if the project exists, False otherwise.
        """
        return project_id in Compose.list()

    @staticmethod
    def _update_projects(include_stopped: bool = False) -> dict[str, list_[str]]:
        """Updates the list of projects by going through containers.

        Params:
            include_stopped: If True, include stopped projects.

        Returns:
            Dict of projects, with the project name as the key and a list of container ids as the value.
        """
        project_container_map = {}
        for container in Containers.list(include_stopped):
            if container.project_id:
                if container.project_id not in project_container_map:
                    project_container_map[container.project_id] = []
                project_container_map[container.project_id].append(container.id)
        return project_container_map


@dataclass
class Volume:
    """Volume class to wrap Docker volumes."""

    name: str
    attrs: dict

    @classmethod
    def from_dict(cls, json_dict: dict) -> "Volume":
        """Create an Image object from a json dictionary.

        Params:
            json_dict: The json dictionary to create the object from.

        Returns:
            The created volume object.
        """
        return cls(
            name=json_dict.get("Name", None),
            attrs=json_dict,
        )

    def remove(self, force: bool = False) -> None:
        """Remove a Docker volume.

        Params:
            force: If True, force the removal of the volume.
        """
        docker = _get_executable("docker")
        try:
            _ = subprocess.run(
                [*docker, "volume", "rm", "--force" if force else "", self.name],
                check=True,
                capture_output=True,
                text=True,
            )
        except subprocess.CalledProcessError as ex:
            raise NotFound(f"Error removing volume {self.name}: {ex}") from ex


class Volumes:
    """Volume class to wrap Docker volumes."""

    @staticmethod
    def create(name: str) -> Volume:
        """Create a Docker volume.

        Params:
            name: The name of the volume to create.

        Returns:
            The created volume object.
        """
        docker = _get_executable("docker")
        try:
            _ = subprocess.run(
                [*docker, "volume", "create", name],
                check=True,
                capture_output=True,
                text=True,
            )
            return Volumes.get(name)
        except subprocess.CalledProcessError as ex:
            raise NotFound(f"Error creating volume {name}: {ex}") from ex

    @staticmethod
    def get(name: str) -> Volume:
        """Get a Docker volume.

        Params:
            name: The name of the volume to get.

        Returns:
            The volume object.
        """
        docker = _get_executable("docker")
        try:
            result = subprocess.run(
                [*docker, "volume", "inspect", name],
                check=True,
                capture_output=True,
                text=True,
            )
            json_dict = json.loads(result.stdout)
        except subprocess.CalledProcessError as ex:
            raise NotFound(f"Volume {name} not found: {ex}") from ex
        if not json_dict:
            raise NotFound(f"Volume {name} not found.")
        return Volume.from_dict(json_dict[0])

    @staticmethod
    def list() -> list[str]:
        """List all Docker volumes.

        Returns:
            List of volume names.
        """
        docker = _get_executable("docker")
        try:
            result = subprocess.run(
                [*docker, "volume", "ls", "-q"],
                check=True,
                capture_output=True,
                text=True,
            )
        except subprocess.CalledProcessError as ex:
            raise APIError(f"Error listing volumes: {ex}") from ex
        return result.stdout.strip().split("\n")


class DockerException(Exception):
    """Base class for Docker CLI exceptions."""

    pass


class BuildError(DockerException):
    """Raised when a build fails."""

    def __init__(self, build_log: list_[str]) -> None:
        self.build_log = build_log

    def __str__(self) -> str:
        return (
            "Docker build failed. Please check the build log for details:\n"
            + "\n".join(self.build_log)
        )


class ContainerError(DockerException):
    """Raised when a container encounters an error."""

    def __init__(
        self,
        container: str | None,
        exit_status: int,
        command: str,
        image: str,
        stderr: bytes,
    ) -> None:
        self.container = container
        self.exit_status = exit_status
        self.command = command
        self.image = image
        self.stderr = stderr


class APIError(DockerException):
    """Raised when a Docker API error occurs."""

    pass


class NotFound(DockerException):
    """Raised when a Docker resource is not found."""

    pass


class ImageNotFound(NotFound):
    """Raised when an image is not found."""

    pass


class CLIDockerClient:
    """Wrapper around Docker CLI to manage Docker containers, images, and projects.

    Initializes a new instance of the current Docker state from the
    perspective of Tesseracts, while mimicking the interface of Docker-Py, with additional
    features for the convenience of Tesseract usage.

    Most calls to CLIDockerClient could be replaced by official Docker-Py Client. However,
    CLIDockerClient by default only sees Tesseract relevant images, containers, and projects;
    the flag `tesseract_only` must be set to False to see non-Tesseract images, containers, and projects.
    CLIDockerClient also has an additional `compose` class for project management that
    Docker-Py does not have due to the Tesseract use case.
    """

    def __init__(self) -> None:
        self.containers = Containers()
        self.images = Images()
        self.compose = Compose()
        self.volumes = Volumes()

    @staticmethod
    def info() -> tuple:
        """Wrapper around docker info call."""
        docker = _get_executable("docker")
        try:
            result = subprocess.run(
                [*docker, "info"],
                check=True,
                capture_output=True,
            )
            return result.stdout, result.stderr
        except subprocess.CalledProcessError as ex:
            raise APIError() from ex


def get_docker_metadata(
    docker_asset_ids: list[str], is_image: bool = False, tesseract_only: bool = True
) -> dict:
    """Get metadata for Docker images/containers.

    Params:
        docker_asset_ids: List of image/container ids to get metadata for.
        is_image: If True, get metadata for images. If False, get metadata for containers.
        tesseract_only: If True, only get metadata for Tesseract images/containers.

    Returns:
        A dict mapping asset ids to their metadata.
    """
    docker = _get_executable("docker")
    if not docker_asset_ids:
        return {}

    # Set metadata in case no images exist and metadata does not get initialized.
    metadata = None
    try:
        result = subprocess.run(
            [*docker, "inspect", *docker_asset_ids],
            check=True,
            capture_output=True,
            text=True,
        )
        metadata = json.loads(result.stdout)

    except subprocess.CalledProcessError as e:
        # Handle the error if some images do not exist.
        error_message = e.stderr
        for asset_id in docker_asset_ids:
            if f"No such image: {asset_id}" in error_message:
                logger.error(f"Image {asset_id} is not a valid image.")
        if "No such object" in error_message:
            raise APIError("Unhealthy container found. Please restart docker.") from e

    if not metadata:
        return {}

    asset_meta_dict = {}
    # Parse the output into a dictionary of only Tesseract assets
    # with the id as the key for easy parsing, and the metadata as the value.
    for asset in metadata:
        env_vars = asset["Config"]["Env"]
        if tesseract_only and (
            not any("TESSERACT_NAME" in env_var for env_var in env_vars)
        ):
            # Do not add items if there is no "TESSERACT_NAME" in env vars.
            continue
        if is_image:
            # If it is an image, use the repotag as the key.
            dict_key = asset["RepoTags"]
            if not dict_key:
                # Old dangling images do not have RepoTags.
                continue
            dict_key = dict_key[0]
        else:
            dict_key = asset["Id"][:12]
        asset_meta_dict[dict_key] = asset
    return asset_meta_dict


def build_docker_image(
    path: str | Path,
    tag: str,
    dockerfile: str | Path,
    inject_ssh: bool = False,
    print_and_exit: bool = False,
) -> Image | None:
    """Build a Docker image from a Dockerfile using BuildKit.

    Params:
        path: Path to the directory containing the Dockerfile.
        tag: The name of the image to build.
        dockerfile: path within the build context to the Dockerfile.
        inject_ssh: If True, inject SSH keys into the build.
        print_and_exit: If True, log the build command and exit without building.

    Returns:
        Built Image object if print_and_exit is False, otherwise None.
    """
    # use an instantiated client here, which may be mocked in tests
    client = CLIDockerClient()
    build_args = dict(path=path, tag=tag, dockerfile=dockerfile)

    if inject_ssh:
        ssh_sock = os.environ.get("SSH_AUTH_SOCK")
        if ssh_sock is None:
            raise ValueError(
                "SSH_AUTH_SOCK environment variable not set (try running `ssh-agent`)"
            )

        ssh_keys = subprocess.run(["ssh-add", "-L"], capture_output=True)
        if ssh_keys.returncode != 0 or not ssh_keys.stdout:
            raise ValueError("No SSH keys found in SSH agent (try running `ssh-add`)")
        build_args["ssh"] = f"default={ssh_sock}"

    build_cmd = Images._get_buildx_command(**build_args)

    if print_and_exit:
        logger.info(
            f"To build the Docker image manually, run:\n    $ {shlex.join(build_cmd)}"
        )
        return None

    return client.images.buildx(**build_args)



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""An instantiated version of the CLI app with user commands added.

Used only for generating documentation. As such, we perform some additional
formatting on the docstrings to make them more readable.

!! Do not use for anything else !!
"""

import copy
from textwrap import indent

from tesseract_core.runtime.cli import (
    _add_user_commands_to_cli,
    _prettify_docstring,
    tesseract_runtime,
)
from tesseract_core.runtime.core import create_endpoints, get_tesseract_api

tesseract_api = get_tesseract_api()

# purge dummy docstrings so they don't leak into the docs
for func in tesseract_api.__dict__.values():
    # don't touch non-functions or functions from other modules
    if not callable(func) or func.__module__ != "tesseract_api":
        continue
    func.__doc__ = None

endpoints = create_endpoints(tesseract_api)

# format docstrings to play well with autodocs
for func in endpoints:
    docstring_parts = [_prettify_docstring(func.__doc__)]

    # populate endpoint docstrings with field info
    input_schema = func.__annotations__.get("payload")
    input_docs = []
    if hasattr(input_schema, "model_fields"):
        for field_name, field in input_schema.model_fields.items():
            input_docs.append(
                f"{field_name} ({field.annotation.__name__}): {field.description}"
            )
    if input_docs:
        docstring_parts.append("")
        docstring_parts.append("Arguments:")
        docstring_parts.append(indent("\n".join(input_docs), "  "))

    output_schema = func.__annotations__.get("return")
    output_docs = []
    if hasattr(output_schema, "model_fields"):
        for field_name, field in output_schema.model_fields.items():
            output_docs.append(
                f"{field_name} ({field.annotation.__name__}): {field.description}"
            )
    if output_docs:
        docstring_parts.append("")
        docstring_parts.append("Returns:")
        docstring_parts.append(indent("\n".join(output_docs), "  "))

    func.__doc__ = "\n".join(docstring_parts)

# add all user-defined functions to the global namespace, so we can do
# `from tesseract_core.runtime.app_cli import jacobian`
globals().update({func.__name__: func for func in endpoints})

tesseract_runtime_cli = copy.deepcopy(tesseract_runtime)
_add_user_commands_to_cli(tesseract_runtime_cli, out_stream=None)



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""End-to-end tests for Tesseract workflows."""

import json
import os
import subprocess
from pathlib import Path

import pytest
import requests
import yaml
from common import build_tesseract, image_exists
from typer.testing import CliRunner

from tesseract_core.sdk.cli import AVAILABLE_RECIPES, app
from tesseract_core.sdk.docker_client import is_podman


@pytest.fixture(scope="module")
def built_image_name(
    docker_client,
    docker_cleanup_module,
    shared_dummy_image_name,
    dummy_tesseract_location,
):
    """Build the dummy Tesseract image for the tests."""
    image_name = build_tesseract(
        docker_client, dummy_tesseract_location, shared_dummy_image_name
    )
    assert image_exists(docker_client, image_name)
    docker_cleanup_module["images"].append(image_name)
    yield image_name


tested_images = ("ubuntu:24.04",)

build_matrix = [
    *[(tag, None, None) for tag in (True, False)],
    *[(False, r, None) for r in AVAILABLE_RECIPES],
    *[(False, None, img) for img in tested_images],
]


@pytest.mark.parametrize("tag,recipe,base_image", build_matrix)
def test_build_from_init_endtoend(
    docker_client, docker_cleanup, dummy_image_name, tmp_path, tag, recipe, base_image
):
    """Test that a trivial (empty) Tesseract image can be built from init."""
    cli_runner = CliRunner(mix_stderr=False)

    init_args = ["init", "--target-dir", str(tmp_path), "--name", dummy_image_name]
    if recipe:
        init_args.extend(["--recipe", recipe])

    result = cli_runner.invoke(app, init_args, catch_exceptions=False)
    assert result.exit_code == 0, result.stderr
    assert (tmp_path / "tesseract_api.py").exists()
    with open(tmp_path / "tesseract_config.yaml") as config_yaml:
        assert yaml.safe_load(config_yaml)["name"] == dummy_image_name

    img_tag = "foo" if tag else None

    config_override = {}
    if base_image is not None:
        config_override["build_config.base_image"] = base_image

    image_name = build_tesseract(
        docker_client,
        tmp_path,
        dummy_image_name,
        config_override=config_override,
        tag=img_tag,
    )

    docker_cleanup["images"].append(image_name)
    assert image_exists(docker_client, image_name)

    # Test that the image can be run and that --help is forwarded correctly
    result = cli_runner.invoke(
        app,
        [
            "run",
            image_name,
            "apply",
            "--help",
        ],
        catch_exceptions=False,
    )
    assert f"Usage: tesseract run {image_name} apply" in result.stderr


@pytest.mark.parametrize("skip_checks", [True, False])
def test_build_generate_only(dummy_tesseract_location, skip_checks):
    """Test output of build with --generate_only flag."""
    cli_runner = CliRunner(mix_stderr=False)
    build_res = cli_runner.invoke(
        app,
        [
            "build",
            str(dummy_tesseract_location),
            "--generate-only",
            *(
                ("--config-override=build_config.skip_checks=True",)
                if skip_checks
                else ()
            ),
        ],
        # Ensure that the output is not truncated
        env={"COLUMNS": "1000"},
        catch_exceptions=False,
    )
    assert build_res.exit_code == 0, build_res.stderr
    # Check that stdout contains build command
    command = "buildx build"
    assert command in build_res.stderr

    build_dir = Path(build_res.stdout.strip())
    assert build_dir.exists()
    dockerfile_path = build_dir / "Dockerfile"
    assert dockerfile_path.exists()

    with open(build_dir / "Dockerfile") as f:
        docker_file_contents = f.read()
        if skip_checks:
            assert 'RUN ["tesseract-runtime", "check"]' not in docker_file_contents
        else:
            assert 'RUN ["tesseract-runtime", "check"]' in docker_file_contents


@pytest.mark.parametrize("no_compose", [True, False])
def test_env_passthrough_serve(
    docker_cleanup, docker_client, built_image_name, no_compose
):
    """Ensure we can pass environment variables to tesseracts when serving."""
    run_res = subprocess.run(
        [
            "tesseract",
            "serve",
            built_image_name,
            "--env=TEST_ENV_VAR=foo",
            *(["--no-compose"] if no_compose else []),
        ],
        capture_output=True,
        text=True,
    )
    assert run_res.returncode == 0, run_res.stderr
    assert run_res.stdout

    project_meta = json.loads(run_res.stdout)
    project_id = project_meta["project_id"]
    tesseract_id = project_meta["containers"][0]["name"]

    if no_compose:
        docker_cleanup["containers"].append(tesseract_id)
    else:
        docker_cleanup["project_ids"].append(project_id)

    container = docker_client.containers.get(tesseract_id)
    exit_code, output = container.exec_run(["sh", "-c", "echo $TEST_ENV_VAR"])
    assert exit_code == 0, f"Command failed with exit code {exit_code}"
    assert "foo" in output.decode("utf-8"), f"Output was: {output.decode('utf-8')}"


def test_tesseract_list(built_image_name):
    # Test List Command
    cli_runner = CliRunner(mix_stderr=False)

    list_res = cli_runner.invoke(
        app,
        [
            "list",
        ],
        # Ensure that the output is not truncated
        env={"COLUMNS": "1000"},
        catch_exceptions=False,
    )
    assert list_res.exit_code == 0, list_res.stderr
    assert built_image_name.split(":")[0] in list_res.stdout


def test_tesseract_run_stdout(built_image_name):
    # Test List Command
    cli_runner = CliRunner(mix_stderr=False)

    test_commands = ("input-schema", "output-schema", "openapi-schema", "health")

    for command in test_commands:
        run_res = cli_runner.invoke(
            app,
            [
                "run",
                built_image_name,
                command,
            ],
            catch_exceptions=False,
        )
        assert run_res.exit_code == 0, run_res.stderr
        assert run_res.stdout

        try:
            json.loads(run_res.stdout)
        except json.JSONDecodeError:
            print(f"failed to decode {command} stdout as JSON")
            print(run_res.stdout)
            raise


@pytest.mark.parametrize("no_compose", [True, False])
@pytest.mark.parametrize("user", [None, "root", "1000:1000"])
def test_run_as_user(docker_client, built_image_name, user, no_compose, docker_cleanup):
    """Ensure we can run a basic Tesseract image as any user."""
    cli_runner = CliRunner(mix_stderr=False)

    run_res = cli_runner.invoke(
        app,
        [
            "serve",
            built_image_name,
            "--user",
            user,
            *(["--no-compose"] if no_compose else []),
        ],
        catch_exceptions=False,
    )
    assert run_res.exit_code == 0, run_res.stderr

    project_meta = json.loads(run_res.stdout)
    project_id = project_meta["project_id"]
    container = docker_client.containers.get(project_meta["containers"][0]["name"])
    if no_compose:
        docker_cleanup["containers"].append(container)
    else:
        docker_cleanup["project_ids"].append(project_id)

    exit_code, output = container.exec_run(["id", "-u"])
    if user is None:
        expected_user = os.getuid()
    elif user == "root":
        expected_user = 0
    else:
        expected_user = int(user.split(":")[0])

    assert exit_code == 0
    assert output.decode("utf-8").strip() == str(expected_user)


@pytest.mark.parametrize("no_compose", [True, False])
def test_tesseract_serve_pipeline(
    docker_client, built_image_name, no_compose, docker_cleanup
):
    cli_runner = CliRunner(mix_stderr=False)
    project_id = None
    run_res = cli_runner.invoke(
        app,
        [
            "serve",
            built_image_name,
            *(["--no-compose"] if no_compose else []),
        ],
        catch_exceptions=False,
    )

    assert run_res.exit_code == 0, run_res.stderr
    assert run_res.stdout

    project_meta = json.loads(run_res.stdout)

    project_id = project_meta["project_id"]
    if no_compose:
        project_container = docker_client.containers.get(project_id)
        docker_cleanup["containers"].append(project_container)
    else:
        docker_cleanup["project_ids"].append(project_id)
        project_containers = project_meta["containers"][0]["name"]
        if not project_containers:
            raise ValueError(f"Could not find container for project '{project_id}'")

        project_container = docker_client.containers.get(project_containers)

    assert project_container.name == project_meta["containers"][0]["name"]
    assert project_container.host_port == project_meta["containers"][0]["port"]
    assert project_container.host_ip == project_meta["containers"][0]["ip"]

    # Ensure served Tesseract is usable
    res = requests.get(
        f"http://{project_container.host_ip}:{project_container.host_port}/health"
    )
    assert res.status_code == 200, res.text

    # Ensure project id is shown in `tesseract ps`
    run_res = cli_runner.invoke(
        app,
        ["ps"],
        env={"COLUMNS": "1000"},
        catch_exceptions=False,
    )
    assert run_res.exit_code == 0, run_res.stderr
    assert project_id in run_res.stdout
    assert project_container.host_port in run_res.stdout
    assert project_container.host_ip in run_res.stdout
    assert project_container.short_id in run_res.stdout


@pytest.mark.parametrize("tear_all", [True, False])
@pytest.mark.parametrize("no_compose", [True, False])
def test_tesseract_teardown_multiple(built_image_name, tear_all, no_compose):
    """Teardown multiple projects."""
    cli_runner = CliRunner(mix_stderr=False)

    project_ids = []
    try:
        for _ in range(2):
            # Serve
            run_res = cli_runner.invoke(
                app,
                [
                    "serve",
                    built_image_name,
                    *(["--no-compose"] if no_compose else []),
                ],
                catch_exceptions=False,
            )
            assert run_res.exit_code == 0, run_res.stderr
            assert run_res.stdout

            project_meta = json.loads(run_res.stdout)

            project_id = project_meta["project_id"]
            project_ids.append(project_id)

    finally:
        # Teardown multiple/all
        args = ["teardown"]
        if tear_all:
            args.extend(["--all"])
        else:
            args.extend(project_ids)

        run_res = cli_runner.invoke(
            app,
            args,
            catch_exceptions=False,
        )
        assert run_res.exit_code == 0, run_res.stderr
        # Ensure all projects are killed
        run_res = cli_runner.invoke(
            app,
            ["ps"],
            env={"COLUMNS": "1000"},
            catch_exceptions=False,
        )
        assert run_res.exit_code == 0, run_res.stderr
        for project_id in project_ids:
            assert project_id not in run_res.stdout


def test_tesseract_serve_ports_error(built_image_name):
    """Check error handling for serve -p flag."""
    cli_runner = CliRunner(mix_stderr=False)

    # Check multiple Tesseracts being served.
    run_res = cli_runner.invoke(
        app,
        [
            "serve",
            built_image_name,
            built_image_name,
            built_image_name,
            "-p",
            "8000-8001",
        ],
        env={"COLUMNS": "1000"},
        catch_exceptions=False,
    )
    assert run_res.exit_code
    assert (
        "Port specification only works if exactly one Tesseract is being served."
        in run_res.stderr
    )

    # Check invalid ports.
    run_res = cli_runner.invoke(
        app,
        [
            "serve",
            built_image_name,
            "-p",
            "8000-999999",
        ],
        env={"COLUMNS": "1000"},
        catch_exceptions=False,
    )
    assert run_res.exit_code
    assert "Ports '8000-999999' must be between" in run_res.stderr

    # Check poorly formatted ports.
    run_res = cli_runner.invoke(
        app,
        [
            "serve",
            built_image_name,
            "-p",
            "8000:8081",
        ],
        env={"COLUMNS": "1000"},
        catch_exceptions=False,
    )
    assert run_res.exit_code
    assert "Port '8000:8081' must be single integer or a range" in run_res.stderr

    # Check invalid port range.
    run_res = cli_runner.invoke(
        app,
        [
            "serve",
            built_image_name,
            "-p",
            "8000-7000",
        ],
        env={"COLUMNS": "1000"},
        catch_exceptions=False,
    )
    assert run_res.exit_code
    assert "Start port '8000' must be less than or equal to end" in run_res.stderr


@pytest.mark.parametrize("port", ["fixed", "range"])
def test_tesseract_serve_ports(built_image_name, port, docker_cleanup, free_port):
    """Try to serve multiple Tesseracts on multiple ports."""
    cli_runner = CliRunner(mix_stderr=False)
    project_id = None

    if port == "fixed":
        port_arg = str(free_port)
    elif port == "range":
        port_arg = f"{free_port}-{free_port + 1}"
    else:
        raise ValueError(f"Unknown port type: {port}")

    # Serve tesseract on specified ports.
    run_res = cli_runner.invoke(
        app,
        ["serve", built_image_name, "-p", port_arg],
        catch_exceptions=False,
    )
    assert run_res.exit_code == 0, run_res.stderr
    assert run_res.stdout

    project_meta = json.loads(run_res.stdout)
    project_id = project_meta["project_id"]
    docker_cleanup["project_ids"].append(project_id)

    # Ensure that actual used ports are in the specified port range.
    test_ports = port_arg.split("-")
    start_port = int(test_ports[0])
    end_port = int(test_ports[1]) if len(test_ports) > 1 else start_port

    actual_port = int(project_meta["containers"][0]["port"])
    assert actual_port in range(start_port, end_port + 1)

    # Ensure specified ports are in `tesseract ps` and served Tesseracts are usable.
    run_res = cli_runner.invoke(
        app,
        ["ps"],
        env={"COLUMNS": "1000"},
        catch_exceptions=False,
    )

    res = requests.get(f"http://localhost:{actual_port}/health")
    assert res.status_code == 200, res.text
    assert str(actual_port) in run_res.stdout


@pytest.mark.parametrize("no_compose", [True, False])
@pytest.mark.parametrize("volume_type", ["bind", "named"])
@pytest.mark.parametrize("user", [None, "root", "1000:1000"])
def test_tesseract_serve_volume_permissions(
    built_image_name,
    docker_client,
    docker_volume,
    tmp_path,
    docker_cleanup,
    user,
    volume_type,
    no_compose,
):
    """Test serving Tesseract with a Docker volume or bind mount.

    This should cover most permissions issues that can arise with Docker volumes.
    """
    if is_podman() and not no_compose:
        pytest.xfail("Podman does not support --no-compose option.")

    cli_runner = CliRunner(mix_stderr=False)
    project_id = None

    dest = Path("/tesseract/output_data")

    if volume_type == "bind":
        # Use bind mount with a temporary directory
        volume_to_bind = str(tmp_path)
    elif volume_type == "named":
        # Use docker volume instead of bind mounting
        volume_to_bind = docker_volume.name
    else:
        raise ValueError(f"Unknown volume type: {volume_type}")

    run_res = cli_runner.invoke(
        app,
        [
            "serve",
            "--volume",
            f"{volume_to_bind}:{dest}:rw",
            *(("--user", user) if user else []),
            built_image_name,
            *(("--no-compose",) if no_compose else [built_image_name]),
        ],
        catch_exceptions=False,
    )
    assert run_res.exit_code == 0, run_res.stderr
    assert run_res.stdout

    project_meta = json.loads(run_res.stdout)
    project_id = project_meta["project_id"]

    if no_compose:
        docker_cleanup["containers"].append(project_id)
    else:
        docker_cleanup["project_ids"].append(project_id)

    tesseract0_id = project_meta["containers"][0]["name"]
    tesseract0 = docker_client.containers.get(tesseract0_id)

    # Sanity check: Should always be allowed to read/write files in the default workdir
    exit_code, output = tesseract0.exec_run(["touch", "./test.txt"])
    assert exit_code == 0, output.decode()

    if volume_type == "bind":
        # Create file outside the containers and check it from inside the container
        tmpfile = Path(tmp_path) / "hi"
        with open(tmpfile, "w") as hello:
            hello.write("world")
            hello.flush()

        if volume_type == "bind" and user not in (None, "root"):
            # If we are not running as root, ensure the file is readable by the target user
            tmp_path.chmod(0o777)
            tmpfile.chmod(0o644)

        exit_code, output = tesseract0.exec_run(["cat", f"{dest}/hi"])
        assert exit_code == 0
        assert output.decode() == "world"

    # Create file inside a container and access it from the other
    bar_file = dest / "bar"
    exit_code, output = tesseract0.exec_run(["ls", "-la", str(dest)])
    assert exit_code == 0, output.decode()

    exit_code, output = tesseract0.exec_run(["touch", str(bar_file)])
    assert exit_code == 0

    if not no_compose:
        tesseract1_id = project_meta["containers"][1]["name"]
        tesseract1 = docker_client.containers.get(tesseract1_id)

        exit_code, output = tesseract1.exec_run(["cat", str(bar_file)])
        assert exit_code == 0
        exit_code, output = tesseract1.exec_run(
            ["bash", "-c", f'echo "hello" > {bar_file}']
        )
        assert exit_code == 0

    if volume_type == "bind":
        # The file should exist outside the container
        assert (tmp_path / "bar").exists()


def test_tesseract_serve_interop(built_image_name, docker_client, docker_cleanup):
    cli_runner = CliRunner(mix_stderr=False)

    run_res = cli_runner.invoke(
        app,
        [
            "serve",
            built_image_name,
            built_image_name,
            "--service-names",
            "tess-1,tess-2",
        ],
        env={"COLUMNS": "1000"},
        catch_exceptions=False,
    )
    assert run_res.exit_code == 0

    project_meta = json.loads(run_res.stdout)
    project_id = project_meta["project_id"]
    docker_cleanup["project_ids"].append(project_id)

    project_containers = [project_meta["containers"][i]["name"] for i in range(2)]

    tess_1 = docker_client.containers.get(project_containers[0])

    returncode, stdout = tess_1.exec_run(
        [
            "python",
            "-c",
            'import requests; requests.get("http://tess-2:8000/health").raise_for_status()',
        ]
    )
    assert returncode == 0, stdout.decode()


@pytest.mark.parametrize("no_compose", [True, False])
def test_serve_nonstandard_host_ip(
    docker_client, built_image_name, docker_cleanup, free_port, no_compose
):
    """Test serving Tesseract with a non-standard host IP."""

    def _get_host_ip():
        """Get a network interface IP address that is not localhost."""
        import socket
        from contextlib import closing

        with closing(socket.socket(socket.AF_INET, socket.SOCK_DGRAM)) as s:
            # We ping to the Google DNS server to get a valid external IP address
            s.connect(("8.8.8.8", 80))
            return s.getsockname()[0]

    cli_runner = CliRunner(mix_stderr=False)
    project_id = None

    # Use a non-standard host IP
    host_ip = _get_host_ip()
    assert host_ip not in ("", "127.0.0.1", "localhost")  # sanity check

    run_res = cli_runner.invoke(
        app,
        [
            "serve",
            built_image_name,
            "-p",
            str(free_port),
            "--host-ip",
            host_ip,
            *(["--no-compose"] if no_compose else []),
        ],
        catch_exceptions=False,
    )
    assert run_res.exit_code == 0, run_res.stderr
    assert run_res.stdout
    project_meta = json.loads(run_res.stdout)
    project_id = project_meta["project_id"]

    if no_compose:
        docker_cleanup["containers"].append(project_id)
    else:
        docker_cleanup["project_ids"].append(project_id)

    project_container = docker_client.containers.get(
        project_meta["containers"][0]["name"]
    )
    assert project_container.host_ip == host_ip

    res = requests.get(f"http://{host_ip}:{project_container.host_port}/health")
    assert res.status_code == 200, res.text

    with pytest.raises(requests.ConnectionError):
        # Ensure that the Tesseract is not accessible from localhost
        requests.get(f"http://localhost:{project_container.host_port}/health")


def test_tesseract_cli_options_parsing(built_image_name, tmpdir):
    cli_runner = CliRunner(mix_stderr=False)

    examples_dir = Path(__file__).parent.parent.parent / "examples"
    example_inputs = examples_dir / "vectoradd" / "example_inputs.json"

    test_commands = (
        ["apply", "-f", "json+binref", "-o", str(tmpdir), f"@{example_inputs}"],
        ["apply", f"@{example_inputs}", "-f", "json+binref", "-o", str(tmpdir)],
        ["apply", "-o", str(tmpdir), f"@{example_inputs}", "-f", "json+binref"],
    )

    for args in test_commands:
        run_res = cli_runner.invoke(
            app,
            [
                "run",
                built_image_name,
                *args,
            ],
            catch_exceptions=False,
        )
        assert run_res.exit_code == 0, run_res.stderr

        with open(Path(tmpdir) / "results.json") as fi:
            results = fi.read()
            assert ".bin:0" in results


def test_tarball_install(dummy_tesseract_package, docker_cleanup):
    import subprocess
    from textwrap import dedent

    tesseract_api = dedent(
        """
    import cowsay
    from pydantic import BaseModel

    class InputSchema(BaseModel):
        message: str = "Hello, Tesseractor!"

    class OutputSchema(BaseModel):
        out: str

    def apply(inputs: InputSchema) -> OutputSchema:
        return OutputSchema(out=cowsay.get_output_string("cow", inputs.message))
    """
    )

    tesseract_requirements = "./cowsay-6.1-py3-none-any.whl"

    subprocess.run(
        ["pip", "download", "cowsay==6.1", "-d", str(dummy_tesseract_package)]
    )
    with open(dummy_tesseract_package / "tesseract_api.py", "w") as f:
        f.write(tesseract_api)
    with open(dummy_tesseract_package / "tesseract_requirements.txt", "w") as f:
        f.write(tesseract_requirements)

    cli_runner = CliRunner(mix_stderr=False)
    result = cli_runner.invoke(
        app,
        ["--loglevel", "debug", "build", str(dummy_tesseract_package)],
        catch_exceptions=False,
    )
    assert result.exit_code == 0, result.stderr

    img_tag = json.loads(result.stdout)[0]
    docker_cleanup["images"].append(img_tag)



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from typing import Self

import numpy as np
from pydantic import BaseModel, Field, model_validator

from tesseract_core.runtime import Array, Differentiable, Float32, Float64


class InputSchema(BaseModel):
    a: Differentiable[Array[(None,), Float32]] = Field(
        description="An arbitrary vector."
    )
    b: Differentiable[Array[(None,), Float32]] = Field(
        description="An arbitrary vector. Needs to have the same dimensions as a."
    )
    s: Differentiable[Float32] = Field(description="A scalar.", default=1)
    normalize: bool = Field(
        description="True if the output should be normalized, False otherwise.",
        default=False,
    )

    @model_validator(mode="after")
    def validate_shape_inputs(self) -> Self:
        if self.a.shape != self.b.shape:
            raise ValueError(
                f"a and b must have the same shape. "
                f"Got {self.a.shape} and {self.b.shape} instead."
            )
        return self


class OutputSchema(BaseModel):
    result: Differentiable[Array[(None,), Float64]] = Field(
        description="Vector s·a + b"
    )


def apply(inputs: InputSchema) -> OutputSchema:
    """Multiplies a vector `a` by `s`, and sums the result to `b`."""
    result = inputs.a * inputs.s + inputs.b

    if inputs.normalize:
        norm = np.linalg.norm(result, ord=2)
        result /= norm

    return OutputSchema(result=result)


def jacobian(
    inputs: InputSchema,
    jac_inputs: set[str],
    jac_outputs: set[str],
):
    assert jac_outputs == {"result"}
    n = len(inputs.a)

    partials = {}
    partials["a"] = np.eye(n) * inputs.s
    partials["b"] = np.eye(n)
    partials["s"] = inputs.a

    if inputs.normalize:
        result = inputs.a * inputs.s + inputs.b
        norm = np.linalg.norm(result, ord=2)
        partials["a"] = (
            partials["a"] / norm
            - np.outer(result, (inputs.a + inputs.s * inputs.b)) / norm**3
        )
        partials["b"] = (
            partials["b"] / norm
            - np.outer(result, (inputs.s * inputs.a + inputs.b)) / norm**3
        )
        partials["s"] = partials["s"] - (inputs.a + inputs.s * inputs.b) / norm**3 * (
            inputs.s * inputs.a * inputs.a + 2 * inputs.a * inputs.b
        )

    jacobian = {"result": {v: partials[v] for v in jac_inputs}}
    return jacobian
